# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

from libc.time cimport tm, time_t
from libc.string cimport memset
from libc.stdint cimport int64_t, int32_t, int16_t, uint8_t, int8_t
from cpython.bytes cimport PyBytes_AS_STRING

import datetime

cdef extern from "time.h":
    time_t timegm(tm *timeptr)

cdef const int64_t MICROS_PER_SEC = 1_000_000
cdef int64_t MICROS_PER_MIN = 60 * MICROS_PER_SEC
cdef int64_t MICROS_PER_HOUR = 60 * MICROS_PER_MIN
cdef int64_t MICROS_PER_DAY = 24 * MICROS_PER_HOUR
cdef const unsigned char CHAR_SPACE = 32
cdef const unsigned char CHAR_T = 84
cdef const unsigned char CHAR_COLON = 58
cdef const unsigned char CHAR_DOT = 46
cdef const unsigned char CHAR_Z = 90
cdef const unsigned char CHAR_PLUS = 43
cdef const unsigned char CHAR_MINUS = 45
cdef const unsigned char CHAR_0 = 48
cdef const unsigned char CHAR_9 = 57

cdef struct ParsedDateTime:
    int16_t year
    uint8_t month
    uint8_t day
    uint8_t hour
    uint8_t minute
    uint8_t second
    int32_t micros
    int8_t offset_sign
    uint8_t offset_hour
    uint8_t offset_minute

cdef inline int parse_2digit(const char* s):
    return (s[0] - CHAR_0) * 10 + (s[1] - CHAR_0)

cdef inline int parse_4digit(const char* s):
    return (s[0] - CHAR_0) * 1000 + (s[1] - CHAR_0) * 100 + (s[2] - CHAR_0) * 10 + (s[3] - CHAR_0)

cdef inline bint is_digit(char c):
    return CHAR_0 <= c <= CHAR_9

cdef inline bint is_digit_range(const char* s, Py_ssize_t start, Py_ssize_t end) nogil:
    cdef const char* ptr = s + start
    cdef Py_ssize_t i = end - start
    while i > 0:
        if ptr[0] < CHAR_0 or ptr[0] > CHAR_9:
            return False
        ptr += 1
        i -= 1
    return True

cdef inline int days_in_month(int year, int month):
    if month == 2:
        if (year % 4 == 0 and year % 100 != 0) or (year % 400 == 0):
            return 29
        return 28
    elif month in (1, 3, 5, 7, 8, 10, 12):
        return 31
    else:
        return 30


cdef inline ParsedDateTime _parse_iso_parts(bytes bts):
    cdef Py_ssize_t n = len(bts)
    cdef const char* s = PyBytes_AS_STRING(bts)
    cdef ParsedDateTime out
    cdef int i = 0
    cdef int digit_count = 0

    out.year = parse_4digit(s)
    out.month = parse_2digit(s + 5)
    out.day = parse_2digit(s + 8)
    out.hour = 0
    out.minute = 0
    out.second = 0
    out.micros = 0
    out.offset_sign = 0
    out.offset_hour = 0
    out.offset_minute = 0

    # fewer than 10, more than 26 and the first separators not '-'
    if n < 10 or n > 26 or s[4] != 45 or s[7] != 45:
        raise ValueError("Invalid ISO timestamp")

    if n == 10:
        if not (1 <= out.month <= 12):
            raise ValueError("Month must be between 1 and 12")
        if not (1 <= out.day <= days_in_month(out.year, out.month)):
            raise ValueError("Invalid day for given month/year")
        return out

    if n >= 16 and (s[10] == CHAR_T or s[10] == CHAR_SPACE):
        out.hour = parse_2digit(s + 11)
        out.minute = parse_2digit(s + 14)
        i = 16

        if n >= 19 and s[16] == CHAR_COLON:
            out.second = parse_2digit(s + 17)
            i = 19

        if i < n and s[i] == CHAR_DOT:
            i += 1
            while i < n and is_digit(s[i]) and digit_count < 6:
                out.micros = out.micros * 10 + (s[i] - CHAR_0)
                i += 1
                digit_count += 1
            while digit_count < 6:
                out.micros *= 10
                digit_count += 1

        if i < n:
            if s[i] == CHAR_Z:
                i += 1
            elif (s[i] == CHAR_PLUS or s[i] == CHAR_MINUS) and (i + 6 <= n):
                out.offset_sign = 1 if s[i] == CHAR_PLUS else -1
                out.offset_hour = parse_2digit(s + i + 1)
                out.offset_minute = parse_2digit(s + i + 4)
                i += 6

        if i != n:
            raise ValueError("Unexpected characters at end of timestamp")
    else:
        raise ValueError("Invalid ISO timestamp")

    # Range checks
    if not (1 <= out.month <= 12):
        raise ValueError("Month must be between 1 and 12")
    if not (1 <= out.day <= days_in_month(out.year, out.month)):
        raise ValueError("Invalid day for given month/year")
    if not (0 <= out.hour < 24):
        raise ValueError("Hour must be between 0 and 23")
    if not (0 <= out.minute < 60):
        raise ValueError("Minute must be between 0 and 59")
    if not (0 <= out.second < 60):
        raise ValueError("Second must be between 0 and 59")

    return out

cpdef int64_t parse_iso_timestamp(bytes bts):
    """
    Parse a ISO format timestamp string to microseconds after the
    linux epoch (including negative values)
    """
    cdef ParsedDateTime p = _parse_iso_parts(bts)
    cdef tm t
    memset(&t, 0, sizeof(tm))
    t.tm_year = p.year - 1900
    t.tm_mon = p.month - 1
    t.tm_mday = p.day
    t.tm_hour = p.hour
    t.tm_min = p.minute
    t.tm_sec = p.second

    cdef time_t epoch = timegm(&t)

    if epoch == -1:
        # Dates before 1970-01-01 need manual handling
        dt = datetime.datetime(p.year, p.month, p.day, p.hour, p.minute, p.second, p.micros)
        delta = dt - datetime.datetime(1970, 1, 1)
        return int(delta.total_seconds() * MICROS_PER_SEC)

    cdef int64_t result = <int64_t>epoch * MICROS_PER_SEC + p.micros
    if p.offset_sign:
        result -= p.offset_sign * (
            (<int64_t>p.offset_hour * 60 + p.offset_minute) * MICROS_PER_MIN
        )
    return result


def parse_iso(bytes bts) -> datetime.datetime:

    cdef ParsedDateTime p

    try:
        p = _parse_iso_parts(bts)
        dt = datetime.datetime(
            p.year, p.month, p.day,
            p.hour, p.minute, p.second,
            p.micros
        )
        if p.offset_sign != 0:
            offset = datetime.timedelta(
                hours=p.offset_sign * p.offset_hour,
                minutes=p.offset_sign * p.offset_minute
            )
            dt = dt - offset  # convert to UTC
        return dt
    except Exception:
        return None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Global Blob Statistics Cache.

This module implements a global statistics cache using an LRU-K2 policy to determine
which items to keep and which to evict. The cache is used to store and retrieve
table or query statistics efficiently, and is shared across all connections and cursors.

The cache is limited by the number of items (MAX_STATISTICS_CACHE_ITEMS), not by memory volume.
Eviction occurs when the item count exceeds the configured maximum.
"""

from typing import Optional

from opteryx.compiled.structures.lru_k import LRU_K
from opteryx.config import MAX_STATISTICS_CACHE_ITEMS
from opteryx.models import RelationStatistics


class _StatsCache:
    """
    Implements a statistics cache using an LRU-K2 eviction policy.
    Stores serialized statistics objects, keyed by bytes.
    """

    slots = "_lru"

    def __init__(self):
        self._lru = LRU_K(k=2)

    def get(self, key: bytes) -> Optional[RelationStatistics]:
        """
        Retrieve a statistics object from the cache by key.
        Returns the deserialized object if found, or None if not present.
        """
        cached_stats = self._lru.get(key)
        if cached_stats is not None:
            return RelationStatistics.from_bytes(cached_stats)
        return None

    def delete(self, key: bytes):
        """
        Remove a statistics object from the cache by key.
        """
        self._lru.delete(key)

    def set(self, key: bytes, value: RelationStatistics):
        """
        Store a statistics object in the cache, serializing it to bytes.
        If the cache exceeds the maximum allowed items, evict the least recently used item.

        Args:
            key: The key associated with the statistics object.
            value: The statistics object to store (will be serialized).
        """
        cached_stats = value.to_bytes()

        # Update LRU cache with the new key and memory pool key if commit succeeds
        self._lru.set(key, cached_stats)
        if self._lru.size > MAX_STATISTICS_CACHE_ITEMS:
            # If the cache size exceeds the limit, evict the least recently used item
            self._lru.evict()

    @property
    def stats(self) -> tuple:
        """
        Return the hit, miss, and eviction statistics for the cache.
        """
        return self._lru.stats

    def __del__(self):
        pass
        # DEBUG: print(f"Statistics Cache <hits={self.stats[0]}, misses={self.stats[1]}, evictions={self.stats[2]}, inserts={self.stats[3]}>")


class StatsCache(_StatsCache):
    """
    Singleton wrapper for the _StatsCache class.
    Ensures only one global statistics cache instance exists.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = cls._create_instance()
        return cls._instance

    @classmethod
    def _create_instance(cls):
        """
        Create a new instance of the underlying _StatsCache.
        """
        return _StatsCache()

    @classmethod
    def reset(cls):
        """
        Reset the StatsCache singleton instance. This is useful when the configuration changes.
        """
        cls._instance = None
        cls._instance = cls._create_instance()



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

import pyarrow
import numpy
cimport numpy
numpy.import_array()

from libc.stdint cimport int32_t, uint8_t, uintptr_t
from cpython.bytes cimport PyBytes_AsString

cdef extern from "string.h":
    int strncasecmp(const char *s1, const char *s2, size_t n)
    int memcmp(const void *s1, const void *s2, size_t n)


cpdef uint8_t[::1] list_allop_neq(object literal, object column):
    """
    Arrow-native equivalent of 'val != ANY(array)' for a column of List<Binary> or List<String>.
    Returns a uint8 mask indicating which rows contain the literal.

    Parameters:
        literal: object
            A string or bytes to compare against.
        column: object
            An Arrow ListArray or ChunkedArray of List<Binary> or List<String>.

    Returns:
        uint8_t[::1]: mask of matching rows.
    """
    cdef:
        list chunks = column if isinstance(column, list) else (column.chunks if hasattr(column, "chunks") else [column])
        Py_ssize_t total_length = 0
        numpy.ndarray[numpy.uint8_t, ndim=1] final_result
        uint8_t[::1] final_view
        Py_ssize_t offset = 0
        uint8_t[::1] chunk_view
        object chunk

    for chunk in chunks:
        total_length += len(chunk)

    final_result = numpy.zeros(total_length, dtype=numpy.uint8)
    final_view = final_result

    offset = 0
    for chunk in chunks:

        column_type = chunk.type
        if not pyarrow.types.is_list(column_type):
            raise TypeError("Expected ListArray or ChunkedArray of lists")
        element_type = column_type.value_type

        if pyarrow.types.is_string(element_type) or pyarrow.types.is_binary(element_type):
            chunk_view = _allop_neq_string_chunk(literal, chunk)
        else:
            chunk_view = _allop_neq_generic_chunk(literal, chunk)

        final_view[offset : offset + len(chunk)] = chunk_view
        offset += len(chunk)

    return final_view


cdef uint8_t[::1] _allop_neq_string_chunk(object literal, object list_array):
    """
    Chunk processor for List<Binary> or List<String>. Uses memcmp.
    """
    cdef:
        bytes literal_bytes = literal.encode('utf-8') if literal is not None else b''
        const char* literal_ptr = PyBytes_AsString(literal_bytes)
        Py_ssize_t literal_len = len(literal_bytes)

        list buffers = list_array.buffers()
        const uint8_t* outer_validity = NULL
        const int32_t* offsets = NULL
        Py_ssize_t arr_offset = list_array.offset
        Py_ssize_t row_count = len(list_array)

        Py_ssize_t offset_in_bits = arr_offset & 7
        Py_ssize_t offset_in_bytes = arr_offset >> 3

        object values_array = list_array.values
        list value_buffers = values_array.buffers()
        const uint8_t* inner_validity = NULL
        const int32_t* value_offsets = NULL
        const char* value_data = NULL

        Py_ssize_t i, j, val_start, val_end, val_len
        Py_ssize_t outer_start, outer_end

        numpy.ndarray[numpy.uint8_t, ndim=1] result = numpy.ones(row_count, dtype=numpy.uint8)
        uint8_t[::1] result_view = result

    # Outer array buffers (list<...>)
    if len(buffers) > 0 and buffers[0]:
        outer_validity = <const uint8_t*> <uintptr_t> buffers[0].address
    if len(buffers) > 1 and buffers[1]:
        offsets = <const int32_t*> <uintptr_t> buffers[1].address

    # Inner value array buffers (flat binary/string values)
    if len(value_buffers) > 0 and value_buffers[0]:
        inner_validity = <const uint8_t*> <uintptr_t> value_buffers[0].address
    if len(value_buffers) > 1 and value_buffers[1]:
        value_offsets = <const int32_t*> <uintptr_t> value_buffers[1].address
    if len(value_buffers) > 2 and value_buffers[2]:
        value_data = <const char*> <uintptr_t> value_buffers[2].address

    for i in range(row_count):
        # Validity check for the outer list
        if outer_validity is not NULL:
            if not (outer_validity[offset_in_bytes + ((offset_in_bits + i) >> 3)] &
                    (1 << ((offset_in_bits + i) & 7))):
                result_view[i] = literal is None
                continue  # Null row → result[i] remains 0

        # Row-level offsets into the flat value array
        outer_start = offsets[arr_offset + i]
        outer_end = offsets[arr_offset + i + 1]

        if outer_end == outer_start:
            result_view[i] = 0
            continue

        for j in range(outer_start, outer_end):
            # NULLs never match, so the condition is None
            if inner_validity is not NULL:
                if not (inner_validity[j >> 3] & (1 << (j & 7))):
                    result_view[i] = 0
                    break  # short-circuit

            val_start = value_offsets[j]
            val_end = value_offsets[j + 1]
            val_len = val_end - val_start

            if val_len != literal_len:
                continue

            if memcmp(value_data + val_start, literal_ptr, <size_t>literal_len) == 0:
                result_view[i] = 0
                break  # short-circuit

    return result_view


cdef uint8_t[::1] _allop_neq_generic_chunk(object literal, object array):
    """
    Generic fallback: compare each value in the array to `literal` using Python equality.
    """
    cdef:
        Py_ssize_t row_count = len(array)
        Py_ssize_t offset = array.offset
        Py_ssize_t i
        object value
        numpy.ndarray[numpy.uint8_t, ndim=1] result = numpy.ones(row_count, dtype=numpy.uint8)
        uint8_t[::1] result_view = result
        bint literal_is_none = literal is None
        bint literal_is_nan = False

    if hasattr(literal, "item"):
        literal = literal.item()
    if hasattr(array, "to_pylist"):
        array = array.to_pylist()

    # Check if literal is NaN
    if isinstance(literal, float) and numpy.isnan(literal):
        literal_is_nan = True

    for i in range(row_count):
        value = array[i + offset]
        if value is None:
            result_view[i] = literal_is_none
        elif len(value) == 0:
            result_view[i] = 0
        elif None in value:
            result_view[i] = 0
        elif literal_is_nan:
            for j in range(len(value)):
                item = value[j]
                if isinstance(item, float) and numpy.isnan(item):
                    result_view[i] = 0  # Found a NaN in array, not all elements ≠ literal
                    break
        elif literal in value:
            result_view[i] = literal_is_none
        elif literal_is_none:
            result_view[i] = 0

    return result_view



# cython: language_level=3
# distutils: language=c++
from libc.stdint cimport uint32_t, uint64_t, int64_t
from libcpp.string cimport string

cdef extern from "Python.h":
    # Correct signature is const, but this was only fixed in Py3.7+
    cdef char* PyUnicode_AsUTF8AndSize(object, Py_ssize_t *)


cdef extern from "util.h":
    cdef void simdjson_error_handler()
    cdef void * flatten_array[T](simd_array src) \
        except +simdjson_error_handler
    cdef void set_active_implementation(Implementation *)


cdef extern from "simdjson.h" namespace "simdjson":
    cdef size_t SIMDJSON_MAXSIZE_BYTES
    cdef size_t SIMDJSON_PADDING
    cdef enum:
        SIMDJSON_VERSION_MAJOR
        SIMDJSON_VERSION_MINOR
        SIMDJSON_VERSION_REVISION

    cdef string minify[T](T) except +simdjson_error_handler

    cdef cppclass Implementation "simdjson::implementation":
        const string &name()
        const string &description()
        bint supported_by_runtime_system() const;

    # This is marked as private and internal, but we don't have a choice.
    cdef cppclass AvailableImplementationList \
            "simdjson::internal::available_implementation_list":
        const Implementation * const *begin()
        const Implementation * const *end()

        const Implementation * operator[](string)

    cdef cppclass atomic_ptr[T]:
        pass

    const AvailableImplementationList& get_available_implementations()
    atomic_ptr["const Implementation"]* get_active_implementation()


cdef extern from "simdjson.h" namespace "simdjson::dom::element_type":
    cdef enum element_type "simdjson::dom::element_type":
        OBJECT,
        ARRAY,
        STRING,
        INT64,
        UINT64,
        DOUBLE,
        BOOL,
        NULL_VALUE


cdef extern from "simdjson.h" namespace "simdjson::dom":
    cdef cppclass simd_array "simdjson::dom::array":
        cppclass iterator:
            iterator()

            iterator operator++()
            bint operator!=(iterator)
            simd_element operator*()

        simd_array.iterator begin()
        simd_array.iterator end()

        size_t size()
        size_t number_of_slots()

        simd_element at(int) except +simdjson_error_handler
        simd_element at_pointer(const char*) except +simdjson_error_handler


    cdef cppclass simd_object "simdjson::dom::object":
        cppclass iterator:
            iterator()

            iterator operator++()
            bint operator!=(iterator)

            uint32_t key_length()
            const char *key_c_str()
            simd_element value()

        simd_object.iterator begin()
        simd_object.iterator end()

        size_t size()

        simd_element at_pointer(const char*) except +simdjson_error_handler
        simd_element operator[](const char*) except +simdjson_error_handler


    cdef cppclass simd_element "simdjson::dom::element":
        element_type type() except +simdjson_error_handler

        const char *get_c_str() except +simdjson_error_handler
        size_t get_string_length() except +simdjson_error_handler

        simd_array get_array() except +simdjson_error_handler
        simd_object get_object() except +simdjson_error_handler
        int64_t get_int64() except +simdjson_error_handler
        uint64_t get_uint64() except +simdjson_error_handler
        double get_double() except +simdjson_error_handler
        bint get_bool() except +simdjson_error_handler


    cdef cppclass simd_parser "simdjson::dom::parser":
        simd_parser() except +simdjson_error_handler
        simd_parser(size_t max_capacity) except +simdjson_error_handler

        simd_element parse(const char *, size_t, bint) \
            except +simdjson_error_handler
        simd_element load(const char *) except +simdjson_error_handler



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy
import pyarrow
from orso.types import OrsoTypes
from pyarrow import compute

from opteryx.compiled import list_ops
from opteryx.datatypes.intervals import MICROSECONDS_PER_DAY
from opteryx.third_party.tktech import csimdjson as simdjson

# Initialize simdjson parser once
parser = simdjson.Parser()


def ArrowOp(documents, elements) -> pyarrow.Array:
    """JSON Selector"""
    element = elements[0]

    # Fast path: if the documents are dicts, delegate to the cython optimized op
    if len(documents) > 0 and isinstance(documents[0], dict):
        return list_ops.cython_arrow_op(documents, element)

    if hasattr(documents, "to_numpy"):
        documents = documents.to_numpy(zero_copy_only=False)

    # Function to extract value from a document
    def extract(doc: bytes, elem: Union[bytes, str]) -> Any:
        value = parser.parse(doc).get(elem)  # type:ignore
        if hasattr(value, "as_list"):
            return value.as_list()
        if hasattr(value, "as_dict"):
            return value.mini
        return value

    try:
        extracted_values = [None if d is None else extract(d, element) for d in documents]
    except ValueError as err:
        from opteryx.exceptions import IncorrectTypeError

        raise IncorrectTypeError("The `->` operator can only be used on JSON documents.") from err

    # Return the result as a PyArrow array
    return pyarrow.array(extracted_values)


def LongArrowOp(documents, elements) -> pyarrow.Array:
    """JSON Selector (as byte string)"""
    element = elements[0]

    if len(documents) > 0 and isinstance(documents[0], dict):
        return list_ops.cython_long_arrow_op(documents, element)

    if hasattr(documents, "to_numpy"):
        documents = documents.to_numpy(zero_copy_only=False)

    def extract(doc: bytes, elem: Union[bytes, str]) -> bytes:
        value = parser.parse(doc).get(elem)  # type:ignore
        if hasattr(value, "mini"):
            return value.mini  # type:ignore
        return None if value is None else str(value).encode()

    try:
        extracted_values = [None if d is None else extract(d, element) for d in documents]
    except ValueError as err:
        from opteryx.exceptions import IncorrectTypeError

        raise IncorrectTypeError("The `->>` operator can only be used on JSON documents.") from err

    # Return the result as a PyArrow array
    return pyarrow.array(extracted_values, type=pyarrow.binary())


def _ip_containment(left: List[Optional[str]], right: List[str]) -> List[Optional[bool]]:
    """
    Check if each IP address in 'left' is contained within the network specified in 'right'.

    Parameters:
        left: List[Optional[str]]
            List of IP addresses as strings.
        right: List[str]
            List containing the network as a string.

    Returns:
        List[Optional[bool]]:
            A list of boolean values indicating if each corresponding IP in 'left' is in 'right'.
    """

    from opteryx.compiled.list_ops import list_ip_in_cidr

    # Normalize the left values to Python str (or None). The compiled
    # Cython routine expects Python str objects; some readers return bytes
    # which cause a TypeError inside the extension. Convert bytes/bytearray
    # and memoryview to str by decoding as utf-8, leave None as-is.
    def _normalize_ip(v):
        if v is None:
            return None
        # memoryview -> bytes
        if isinstance(v, memoryview):
            try:
                v = v.tobytes()
            except Exception:
                v = bytes(v)
        if isinstance(v, (bytes, bytearray)):
            try:
                return v.decode("utf-8")
            except Exception:
                return str(v)
        if not isinstance(v, str):
            return str(v)
        return v

    try:
        normalized_left = [_normalize_ip(v) for v in left]
        # list_ip_in_cidr expects a numpy.ndarray (dtype=object) of Python str
        # objects; ensure we pass the correct type to avoid TypeError.
        import numpy as _numpy

        arr = _numpy.asarray(normalized_left, dtype=object)
        return list_ip_in_cidr(arr, str(right[0]))
    except (IndexError, AttributeError, ValueError, TypeError) as err:
        from opteryx.exceptions import IncorrectTypeError

        raise IncorrectTypeError(
            "The `|` operator can be used as bitwise OR or IP address containment only."
        ) from err


def binary_operations(
    left, left_type: OrsoTypes, operator: str, right, right_type: OrsoTypes
) -> Union[numpy.ndarray, pyarrow.Array]:
    """
    Execute inline operators (e.g. the add in 3 + 4).

    Parameters:
        left: Union[numpy.ndarray, pyarrow.Array]
            The left operand
        operator: str
            The operator to be applied
        right: Union[numpy.ndarray, pyarrow.Array]
            The right operand
    Returns:
        Union[numpy.ndarray, pyarrow.Array]
            The result of the binary operation
    """
    operation = OPERATOR_FUNCTION_MAP.get(operator)

    if operation is None:
        raise NotImplementedError(f"Operator `{operator}` is not implemented!")

    if OrsoTypes.INTERVAL in (left_type, right_type):
        from opteryx.datatypes.intervals import INTERVAL_KERNELS

        function = INTERVAL_KERNELS.get((left_type, right_type, operator))
        if function is None:
            from opteryx.exceptions import UnsupportedTypeError

            raise UnsupportedTypeError(
                f"Cannot perform {operator.upper()} on {left_type} and {right_type}."
            )

        return function(left, left_type, right, right_type, operator)

    if (
        operator == "Minus"
        and left_type in (OrsoTypes.DATE, OrsoTypes.TIMESTAMP)
        and right_type in (OrsoTypes.DATE, OrsoTypes.TIMESTAMP)
    ):
        # substracting dates results in an INTERVAL (months, seconds)
        arr = OPERATOR_FUNCTION_MAP["Minus"](left, right)
        arr64 = arr.astype(numpy.int64)

        # Arrow represents nulls as INT64_MIN
        null_mask = arr64 == numpy.iinfo(numpy.int64).min
        if arr.dtype.name == "timedelta64[D]":
            result = [
                (None if is_null else (0, v * MICROSECONDS_PER_DAY))
                for v, is_null in zip(arr64, null_mask)
            ]
        else:
            result = [(0, v) if not is_null else None for v, is_null in zip(arr64, null_mask)]
        return pyarrow.array(result)

    elif operator == "BitwiseOr" and OrsoTypes.VARCHAR in (left_type, right_type):
        return _ip_containment(left, right)

    elif operator == "StringConcat":
        return compute.binary_join_element_wise(left, right, "")

    return operation(left, right)


# fmt:off
OPERATOR_FUNCTION_MAP: Dict[str, Any] = {
    "Divide": numpy.divide,
    "Minus": numpy.subtract,
    "Modulo": numpy.mod,
    "Multiply": numpy.multiply,
    "Plus": numpy.add,
    "StringConcat": compute.binary_join_element_wise,
    "MyIntegerDivide": lambda left, right: numpy.trunc(numpy.divide(left, right)).astype(numpy.int64),
    "BitwiseOr": numpy.bitwise_or,
    "BitwiseAnd": numpy.bitwise_and,
    "BitwiseXor": numpy.bitwise_xor,
    "ShiftLeft": numpy.left_shift,
    "ShiftRight": numpy.right_shift,
    "Arrow": ArrowOp,
    "LongArrow": LongArrowOp
}

BINARY_OPERATORS = set(OPERATOR_FUNCTION_MAP.keys())

# fmt:on



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Decode files from a raw binary format to a PyArrow Table.
"""

import io
from enum import Enum
from typing import BinaryIO
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import pyarrow
from orso.tools import random_string
from orso.types import OrsoTypes
from pyarrow import parquet

import opteryx.rugo.parquet as parquet_meta
from opteryx.compiled.structures.memory_view_stream import MemoryViewStream
from opteryx.connectors.capabilities import PredicatePushable
from opteryx.exceptions import UnsupportedFileTypeError
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import RelationStatistics
from opteryx.rugo.converters.orso import rugo_to_orso_schema
from opteryx.utils.arrow import post_read_projector


class ExtentionType(str, Enum):
    """labels for the file extentions"""

    DATA = "DATA"
    CONTROL = "CONTROL"


def convert_avro_schema_to_orso_schema(avro_schema):
    from orso.schema import FlatColumn
    from orso.schema import RelationSchema

    avro_to_orso: Dict[str, OrsoTypes] = {
        "long": OrsoTypes.INTEGER,
        "string": OrsoTypes.VARCHAR,
        "timestamp": OrsoTypes.TIMESTAMP,
        "boolean": OrsoTypes.BOOLEAN,
        "array": OrsoTypes.ARRAY,
        "float": OrsoTypes.DOUBLE,
        "double": OrsoTypes.DOUBLE,
        "bytes": OrsoTypes.BLOB,
    }

    columns = []

    for column in avro_schema["fields"]:
        ct = None
        act = column.get("type")
        if isinstance(act, str):
            ct = avro_to_orso.get(act)
        if isinstance(act, list):
            types = [avro_to_orso.get(t) for t in act if t != "null"]
            if len(types) > 0:
                ct = types[0]
        if isinstance(act, dict):
            ct = avro_to_orso.get(act.get("type"))
        fc = FlatColumn(name=column.get("name"), type=ct)
        columns.append(fc)

    return RelationSchema(name=avro_schema.get("name"), columns=columns)


def convert_arrow_schema_to_orso_schema(
    arrow_schema, row_count_metric: Optional[int] = None, row_count_estimate: Optional[int] = None
):
    from orso.schema import FlatColumn
    from orso.schema import RelationSchema

    return RelationSchema(
        name="arrow",
        row_count_metric=row_count_metric,
        row_count_estimate=row_count_estimate,
        columns=[FlatColumn.from_arrow(field) for field in arrow_schema],
    )


def get_decoder(dataset: str) -> Callable:
    """helper routine to get the decoder for a given file"""
    ext = dataset.rpartition(".")[2].lower()
    file_decoder, file_type = KNOWN_EXTENSIONS.get(ext, (None, None))
    if file_type is None:
        raise UnsupportedFileTypeError(f"Unsupported file type: {ext}")
    if file_type != ExtentionType.DATA:  # pragma: no cover
        return do_nothing
    return file_decoder


def do_nothing(buffer: Union[memoryview, bytes], **kwargs):  # pragma: no cover
    """for when you need to look like you're doing something"""
    return None


def filter_records(filters: Optional[list], table: pyarrow.Table) -> pyarrow.Table:
    """
    Apply filters to a PyArrow table that could not be pushed down during the read operation.
    This is a post-read filtering step.

    Parameters:
        filters: Optional[list]
            A list of filter conditions (predicates) to apply to the table.
        table: pyarrow.Table
            The PyArrow table to be filtered.

    Returns:
        pyarrow.Table:
            A new PyArrow table with rows filtered according to the specified conditions.

    Note:
        At this point the columns are the raw column names from the file so we need to ensure
        the filters reference the raw column names not the engine internal 'identity'=
    """
    from opteryx.managers.expression import evaluate
    from opteryx.models import Node

    if isinstance(filters, list) and filters:
        # Create a copy of the filters list to avoid mutating the original.
        filter_copy = [f.copy() for f in filters]
        root = filter_copy.pop()

        # If the left or right side of the root filter node is an identifier, set its identity.
        # This step ensures that the filtering logic aligns with the schema before any renaming.
        if root.left.node_type == NodeType.IDENTIFIER:
            root.left.schema_column.identity = root.left.source_column
        if root.right.node_type == NodeType.IDENTIFIER:
            root.right.schema_column.identity = root.right.source_column

        while filter_copy:
            right = filter_copy.pop()
            if right.left.node_type == NodeType.IDENTIFIER:
                right.left.schema_column.identity = right.left.source_column
            if right.right.node_type == NodeType.IDENTIFIER:
                right.right.schema_column.identity = right.right.source_column
            # Combine the current root with the next filter using an AND node.
            root = Node(
                NodeType.AND,
                left=root,
                right=right,
                schema_column=Node("schema_column", identity=random_string()),
            )
    else:
        root = filters

    mask = evaluate(root, table)
    return table.filter(mask)


def zstd_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    """
    Read zstandard compressed JSONL files
    """
    if just_statistics:
        return None

    from opteryx.third_party.facebook import zstd

    return jsonl_decoder(
        zstd.decompress(buffer),
        projection=projection,
        selection=selection,
        just_schema=just_schema,
    )


def lzma_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    """
    Read lzma compressed JSONL files
    """
    if just_statistics:
        return None

    import lzma

    # similar to zstd path: read bytes and pass decompressed data as memoryview
    # zstandard.open expects a file-like
    if not isinstance(buffer, memoryview):
        buffer = memoryview(buffer)
    buffer = MemoryViewStream(buffer)

    with lzma.open(buffer, "rb") as file:
        decompressed = file.read()
        return jsonl_decoder(
            memoryview(decompressed),
            projection=projection,
            selection=selection,
            just_schema=just_schema,
        )


def parquet_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    force_read: bool = False,
    use_threads: bool = True,
    statistics: Optional[RelationStatistics] = None,
) -> Tuple[int, int, pyarrow.Table]:
    """
    Read parquet formatted files.

    Parameters:
        buffer: Union[memoryview, bytes]
            The input buffer containing the parquet file data.
        projection: List, optional
            List of columns to project.
        selection: optional
            The selection filter.
        just_schema: bool, optional
            Flag to indicate if only schema is needed.
        force_read: bool, optional
            Flag to skip some optimizations.
    Returns:
        Tuple containing number of rows, number of columns, and the table or schema.
    """

    # If it's COUNT(*), we don't need to create a full dataset
    # We have a handler later to sum up the $COUNT(*) column
    # We can use rugo's metadata reader which is faster than pyarrow's
    if projection == [] and selection == []:
        if isinstance(buffer, memoryview):
            metadata = parquet_meta.read_metadata_from_memoryview(
                buffer, include_statistics=False, max_row_groups=1
            )
        else:
            metadata = parquet_meta.read_metadata_from_memoryview(
                memoryview(buffer), include_statistics=False, max_row_groups=1
            )
        num_rows = metadata["num_rows"]
        num_columns = len(metadata["row_groups"][0]["columns"])
        num_bytes = sum(x["total_byte_size"] for x in metadata["row_groups"])
        table = pyarrow.Table.from_arrays([[num_rows]], names=["$COUNT(*)"])
        return (
            num_rows,
            num_columns,
            num_bytes,
            table,
        )

    # Return just the schema if that's all that's needed
    # We can use rugo's metadata reader which is faster than pyarrow's
    if just_schema:
        if isinstance(buffer, memoryview):
            metadata = parquet_meta.read_metadata_from_memoryview(
                buffer, schema_only=True, max_row_groups=1, include_statistics=False
            )
        else:
            metadata = parquet_meta.read_metadata_from_memoryview(
                memoryview(buffer), schema_only=True, max_row_groups=1, include_statistics=False
            )
        return rugo_to_orso_schema(metadata, "parquet")

    # Gather statistics if that's all that's needed
    # We can use rugo's metadata reader which is faster than pyarrow's
    if just_statistics:
        if statistics is None:
            statistics = RelationStatistics()

        if isinstance(buffer, memoryview):
            metadata = parquet_meta.read_metadata_from_memoryview(buffer, include_statistics=True)
        else:
            metadata = parquet_meta.read_metadata_from_memoryview(
                memoryview(buffer), include_statistics=True
            )

        num_rows = metadata["num_rows"]
        statistics.record_count += num_rows

        for row_group in metadata["row_groups"]:
            for column in row_group["columns"]:
                column_name = column["name"]

                min_value = column.get("min")
                if min_value is not None:
                    statistics.update_lower(column_name, min_value)

                max_value = column.get("max")
                if max_value is not None:
                    statistics.update_upper(column_name, max_value)

                null_count = column.get("null_count")
                if null_count:
                    statistics.add_null(column_name, null_count)

        return statistics

    # Use rugo's lightweight metadata reader first (faster than pyarrow)
    if isinstance(buffer, memoryview):
        rmeta = parquet_meta.read_metadata_from_memoryview(buffer)
    else:
        rmeta = parquet_meta.read_metadata_from_memoryview(memoryview(buffer))

    # Build the pieces we need from the rugo metadata
    # schema names (parquet has same columns across row groups usually)
    if rmeta.get("row_groups"):
        schema_names = [c["name"] for c in rmeta["row_groups"][0]["columns"]]
    else:
        schema_names = []

    num_rows = rmeta.get("num_rows")
    # number of columns - try to derive, fallback to length of schema_names
    num_columns = rmeta.get("num_columns") or len(schema_names)

    # total uncompressed size (rugo uses total_byte_size)
    uncompressed_size = sum(
        sum(col.get("total_byte_size", 0) for col in rg.get("columns", []))
        for rg in rmeta.get("row_groups", [])
    )

    # we need to work out if we have a selection which may force us
    # fetching columns just for filtering
    dnf_filter, processed_selection = (
        PredicatePushable.to_dnf(selection) if selection else (None, None)
    )

    # Determine the columns needed for projection and filtering
    projection_set = set(p.source_column for p in projection or [])
    filter_columns = {
        c.value for c in get_all_nodes_of_type(processed_selection, (NodeType.IDENTIFIER,))
    }
    selected_columns = list(projection_set.union(filter_columns).intersection(schema_names))

    # Read all columns if none are selected, unless force_read is set
    if not selected_columns and not force_read:
        selected_columns = []

    # Open the parquet file only once. Fake a file-like object around the buffer
    if isinstance(buffer, memoryview):
        buffer = MemoryViewStream(buffer)

    # Read the parquet table with the optimized column list and selection filters
    table = parquet.read_table(
        buffer,
        columns=selected_columns,
        pre_buffer=True,
        filters=dnf_filter,
        use_threads=use_threads,
        use_pandas_metadata=False,
    )

    # Any filters we couldn't push to PyArrow to read we run here
    if processed_selection:
        table = filter_records(processed_selection, table)

    return (
        num_rows,
        num_columns,
        uncompressed_size,
        table,
    )


def orc_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    """
    Read orc formatted files
    """
    if just_statistics:
        return None

    import pyarrow.orc as orc

    if isinstance(buffer, memoryview):
        # Convert memoryview to bytes for PyArrow BufferReader
        stream = pyarrow.BufferReader(buffer.tobytes())
    elif isinstance(buffer, bytes):
        stream: BinaryIO = io.BytesIO(buffer)
    else:
        stream = buffer

    orc_file = orc.ORCFile(stream)

    if just_schema:
        orc_schema = orc_file.schema
        return convert_arrow_schema_to_orso_schema(orc_schema)

    table = orc_file.read()
    full_shape = table.shape
    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)
    return *full_shape, 0, table


def jsonl_decoder(
    buffer: Union[memoryview, bytes, BinaryIO],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    # rugo is our own library for fast jsonl reading
    from orso.schema import convert_orso_schema_to_arrow_schema

    import opteryx.rugo.jsonl as rj
    from opteryx.rugo.converters.orso import jsonl_to_orso_schema
    from opteryx.utils import count_instances

    if just_statistics:
        return None

    if not isinstance(buffer, memoryview):
        buffer = memoryview(buffer)

    # count newline occurrences in the provided buffer to get the number of rows
    num_rows = count_instances(buffer)

    # If it's COUNT(*), we don't need to create a full dataset
    # We have a handler later to sum up the $COUNT(*) column
    if projection == [] and selection == [] and not just_schema and not just_statistics:
        # Try to use the SIMD-optimized counter from the cython module if available
        table = pyarrow.Table.from_arrays([[num_rows]], names=["$COUNT(*)"])
        return (num_rows, 0, len(buffer), table)

    orso_schema = jsonl_to_orso_schema(rj.get_jsonl_schema(buffer))

    if just_schema:
        return orso_schema

    # Determine the columns needed for projection and filtering
    projection_set = set(p.source_column for p in projection or [])
    filter_columns = {c.value for c in get_all_nodes_of_type(selection, (NodeType.IDENTIFIER,))}
    selected_columns = list(
        projection_set.union(filter_columns).intersection(orso_schema.column_names)
    )

    table = rj.read_jsonl(
        buffer,
        columns=selected_columns,
        parse_objects=False,
    )

    # Convert the returned columns into PyArrow arrays using the Arrow schema
    # derived from the Orso schema. Doing this per-field ensures that Python
    # lists (JSON arrays) are converted into Arrow ListArray types and that
    # bytes/strings/etc. are coerced to the expected Arrow types. If a direct
    arrow_schema = convert_orso_schema_to_arrow_schema(orso_schema)

    arrays = []
    final_fields = []

    for idx, name in enumerate(table["column_names"]):
        field = arrow_schema.field(name)
        column = table["columns"][idx]
        if hasattr(column, "to_arrow"):
            # rugo returns draken vectors; convert to pyarrow arrays
            arrays.append(column.to_arrow())
        else:
            # fallback: convert using pyarrow array constructor
            arrays.append(pyarrow.array(column, type=field.type))
        final_fields.append(field)

    final_schema = pyarrow.schema(final_fields)
    arrow_table = pyarrow.Table.from_arrays(arrays, schema=final_schema)

    if selection:
        arrow_table = filter_records(selection, arrow_table)
    if projection:
        arrow_table = post_read_projector(arrow_table, projection)

    return num_rows, len(table["column_names"]), len(buffer), arrow_table


def csv_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    delimiter: str = ",",
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    if just_statistics:
        return None

    import pyarrow.csv
    from pyarrow.csv import ParseOptions

    if isinstance(buffer, memoryview):
        stream = MemoryViewStream(buffer)
    elif isinstance(buffer, bytes):
        stream: BinaryIO = io.BytesIO(buffer)
    else:
        stream = buffer

    parse_options = ParseOptions(delimiter=delimiter, newlines_in_values=True)
    table = pyarrow.csv.read_csv(stream, parse_options=parse_options)
    schema = table.schema
    if just_schema:
        return convert_arrow_schema_to_orso_schema(schema)

    full_shape = table.shape
    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)

    return *full_shape, 0, table


def tsv_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    return csv_decoder(
        buffer=buffer,
        projection=projection,
        selection=selection,
        delimiter="\t",
        just_statistics=just_statistics,
        just_schema=just_schema,
    )


def psv_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    return csv_decoder(
        buffer=buffer,
        projection=projection,
        selection=selection,
        delimiter="|",
        just_schema=just_schema,
        just_statistics=just_statistics,
    )


def arrow_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    if just_statistics:
        return None

    import pyarrow.feather as pf

    if isinstance(buffer, memoryview):
        stream = MemoryViewStream(buffer)
    elif isinstance(buffer, bytes):
        stream: BinaryIO = io.BytesIO(buffer)
    else:
        stream = buffer

    table = pf.read_table(stream)
    schema = table.schema
    if just_schema:
        return convert_arrow_schema_to_orso_schema(schema)

    full_shape = table.shape
    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)

    return *full_shape, 0, table


def avro_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    """
    AVRO has a number of optimizations to make it faster than a naive implementation;
    the sample test script runs about 7x faster following these changes (the schema
    converter and selecting before convering to pyarrow).

    AVRO is still many many times slower than Parquet - it's not recommended as a
    bulk data format.
    """
    if just_statistics:
        return None

    try:
        import fastavro
    except ImportError:  # pragma: no cover
        from opteryx.exceptions import MissingDependencyError

        raise MissingDependencyError("fastavro")

    if isinstance(buffer, memoryview):
        stream = MemoryViewStream(buffer)
    elif isinstance(buffer, bytes):
        stream: BinaryIO = io.BytesIO(buffer)
    else:
        stream = buffer

    reader = fastavro.reader(stream)

    if just_schema:
        # FastAvro exposes a schema we can convert without reading all the rows
        return convert_avro_schema_to_orso_schema(reader.schema)

    if projection:
        # It's almost always faster to avoid creating the column to convert in arrow
        # than creating and then removing them - although that would probably the fastest step
        projection = {c.value for c in projection}
        table = pyarrow.Table.from_pylist(
            [{k: v for k, v in row.items() if k in projection} for row in reader]
        )
    elif projection == []:
        # Empty table, we don't know the number of rows up front
        table = pyarrow.Table.from_arrays([[0 for r in reader]], ["_"])
    else:
        # Probably never run, convert every row and column to Arrow
        table = pyarrow.Table.from_pylist(list(reader))

    full_shape = table.shape
    if selection:
        # We can't push filters in Fast Avro, so filter here
        table = filter_records(selection, table)

    return *full_shape, 0, table


def ipc_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    if just_statistics:
        return None

    from itertools import chain

    from pyarrow import ipc

    if isinstance(buffer, memoryview):
        stream = MemoryViewStream(buffer)
    elif isinstance(buffer, bytes):
        stream: BinaryIO = io.BytesIO(buffer)
    else:
        stream = buffer

    reader = ipc.open_stream(stream)

    batch_one = next(reader, None)
    if batch_one is None:
        return None

    schema = batch_one.schema

    if just_schema:
        return convert_arrow_schema_to_orso_schema(schema)

    table = pyarrow.Table.from_batches([batch for batch in chain([batch_one], reader)])
    full_shape = table.shape
    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)

    return *full_shape, 0, table


def excel_decoder(
    buffer: Union[memoryview, bytes],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    """
    Reads an Excel file and converts it to a PyArrow table.

    Parameters:
        file_path: str
            Path to the Excel file.
        sheet_name: str, optional
            Name of the sheet to read. If None, reads the first sheet.

    Returns:
        pyarrow.Table
            A PyArrow table containing the Excel data.
    """
    if just_statistics:
        return None

    import pandas

    # Convert buffer to appropriate format for pandas
    if isinstance(buffer, memoryview):
        stream = MemoryViewStream(buffer)
    elif isinstance(buffer, bytes):
        stream: BinaryIO = io.BytesIO(buffer)
    else:
        stream = buffer

    # Read Excel file using pandas
    df = pandas.read_excel(stream)

    # Convert the pandas DataFrame to a PyArrow Table
    table = pyarrow.Table.from_pandas(df)

    if just_schema:
        return convert_arrow_schema_to_orso_schema(table.schema)

    shape = table.shape

    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)

    return *shape, 0, table


def convert_string_view(field: pyarrow.Field) -> pyarrow.Field:
    """
    Recursively replace string_view (type id 39) with string in a pyarrow Field.
    """
    field_type = field.type

    if field_type.id == pyarrow.string_view().id:
        return pyarrow.field(field.name, pyarrow.string(), nullable=field.nullable)

    elif pyarrow.types.is_list(field_type) or pyarrow.types.is_large_list(field_type):
        converted_value_field = convert_string_view(pyarrow.field("item", field_type.value_type))
        if pyarrow.types.is_list(field_type):
            new_type = pyarrow.list_(converted_value_field.type)
        else:
            new_type = pyarrow.large_list(converted_value_field.type)
        return pyarrow.field(field.name, new_type, nullable=field.nullable)

    elif pyarrow.types.is_struct(field_type):
        new_fields = [convert_string_view(subfield) for subfield in field_type]
        new_type = pyarrow.struct(new_fields)
        return pyarrow.field(field.name, new_type, nullable=field.nullable)

    return field


def vortex_decoder(
    buffer: Union[memoryview, bytes, BinaryIO],
    *,
    projection: Optional[list] = None,
    selection: Optional[list] = None,
    just_schema: bool = False,
    just_statistics: bool = False,
    **kwargs,
) -> Tuple[int, int, pyarrow.Table]:
    try:
        import vortex
    except ImportError:
        from opteryx.exceptions import MissingDependencyError

        raise MissingDependencyError("vortex-data")

    import os
    import tempfile

    if just_statistics:
        return None

    # Current version of vortex appears to not be able to read streams
    # this is painfully slow, don't do this.
    # Convert buffer to bytes if needed
    if isinstance(buffer, memoryview):
        buffer_bytes = buffer.tobytes()
    elif isinstance(buffer, bytes):
        buffer_bytes = buffer
    else:
        buffer_bytes = buffer.read()

    with tempfile.NamedTemporaryFile(suffix=".vortex", delete=False) as f:
        f.write(buffer_bytes)
        f.flush()
        tmp_name = f.name
    try:
        table = vortex.open(tmp_name)
    finally:
        os.remove(tmp_name)

    if just_schema:
        arrow_schema = table.to_arrow().schema
        orso_schema = convert_arrow_schema_to_orso_schema(arrow_schema)
        return orso_schema

    # If it's COUNT(*), we don't need to create a full dataset
    # We have a handler later to sum up the $COUNT(*) column
    if projection == [] and selection == [] and not just_schema and not just_statistics:
        num_rows = len(table)
        table = pyarrow.Table.from_arrays([[num_rows]], names=["$COUNT(*)"])
        return (num_rows, 0, 0, table)

    # we currently aren't pushing filters into vortex so we need to read
    # the columns we're filtering by
    projection_set = set(p.source_column for p in projection or [])
    filter_columns = {c.value for c in get_all_nodes_of_type(selection, (NodeType.IDENTIFIER,))}
    selected_columns = list(projection_set.union(filter_columns))

    # convert to pyarrow table
    table = table.to_arrow(projection=selected_columns).read_all()
    shape = table.shape

    # string views aren't properly supported - only ever seen them in vortex files
    new_schema = pyarrow.schema([convert_string_view(field) for field in table.schema])
    table = table.cast(new_schema)

    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)

    return *shape, 0, table


# for types we know about, set up how we handle them
KNOWN_EXTENSIONS: Dict[str, Tuple[Callable, str]] = {
    "avro": (avro_decoder, ExtentionType.DATA),
    "complete": (do_nothing, ExtentionType.CONTROL),
    "manifest": (do_nothing, ExtentionType.CONTROL),
    "ignore": (do_nothing, ExtentionType.CONTROL),
    "arrow": (arrow_decoder, ExtentionType.DATA),  # feather
    "csv": (csv_decoder, ExtentionType.DATA),
    "ipc": (ipc_decoder, ExtentionType.DATA),
    "jsonl": (jsonl_decoder, ExtentionType.DATA),
    "orc": (orc_decoder, ExtentionType.DATA),
    "parquet": (parquet_decoder, ExtentionType.DATA),
    "tsv": (tsv_decoder, ExtentionType.DATA),
    "psv": (psv_decoder, ExtentionType.DATA),
    "zstd": (zstd_decoder, ExtentionType.DATA),  # jsonl/zstd
    "lzma": (lzma_decoder, ExtentionType.DATA),  # jsonl/lzma
    "vortex": (vortex_decoder, ExtentionType.DATA),
    "xlsx": (excel_decoder, ExtentionType.DATA),  # jsonl/lzma
}

VALID_EXTENSIONS = set(f".{ext}" for ext in KNOWN_EXTENSIONS)
TUPLE_OF_VALID_EXTENSIONS = tuple(VALID_EXTENSIONS)
DATA_EXTENSIONS = set(
    f".{ext}" for ext, conf in KNOWN_EXTENSIONS.items() if conf[1] == ExtentionType.DATA
)



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: overflowcheck=False
# cython: lintrule=ignore

from libc.stdint cimport int64_t, uint64_t, int32_t
from libc.stddef cimport size_t
from libcpp.pair cimport pair
from libcpp.vector cimport vector

# Identity Hash Definition - not part of abseil but used by our implementation
# We prehash the values before putting them into the Map & Set, so don't rehash
cdef extern from "identity_hash.h":
    cdef cppclass IdentityHash:
        size_t operator()(uint64_t value) const


cdef extern from "absl/container/flat_hash_map.h" namespace "absl" nogil:
    cdef cppclass flat_hash_map[K, V, HashFunc]:
        flat_hash_map()
        V& operator[](K key)
        size_t size() const
        void clear()

cdef class FlatHashMap:
    cdef flat_hash_map[uint64_t, vector[int64_t], IdentityHash] _map

    cpdef insert(self, uint64_t key, int64_t value)
    cpdef size_t size(self)
    cpdef clear(self)
    cpdef vector[int64_t] get(self, uint64_t key)

cdef extern from "absl/container/flat_hash_set.h" namespace "absl" nogil:
    cdef cppclass flat_hash_set[T, HashFunc]:
        flat_hash_set()
        pair[long, bint] insert(T value)
        size_t size() const
        bint contains(T value) const
        void reserve(int64_t value)

cdef extern from "flat_hash_set_helpers.h" namespace "opteryx" nogil:
    void flat_hash_set_insert_many(
        flat_hash_set[uint64_t, IdentityHash]& target,
        const uint64_t* values,
        size_t length,
    )

cdef class FlatHashSet:
    cdef flat_hash_set[uint64_t, IdentityHash] _set

    cdef inline bint insert(self, uint64_t value) noexcept nogil
    cdef inline void just_insert(self, uint64_t value) noexcept nogil
    cdef inline void insert_many(self, uint64_t* values, Py_ssize_t length) noexcept nogil
    cdef inline size_t size(self) noexcept nogil
    cdef inline bint contains(self, uint64_t value) noexcept nogil
    cdef inline void reserve(self, int64_t capacity) noexcept nogil
    cdef vector[int64_t] find_new_indices(self, uint64_t* hashes, Py_ssize_t length) noexcept nogil
    cdef Py_ssize_t find_new_indices_out(self, uint64_t* hashes, Py_ssize_t length, int64_t* out_indices) noexcept nogil
    cdef Py_ssize_t find_new_indices_out_32(self, uint64_t* hashes, Py_ssize_t length, int32_t* out_indices) noexcept nogil

    cpdef size_t items(self)



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

from libc.stdint cimport uint8_t, uint32_t, uint64_t, int64_t
cimport numpy

# Declaration of the BloomFilter class
cdef class BloomFilter:
    cdef uint64_t* bit_array
    cdef uint32_t bit64_array_size
    cdef uint32_t bit_array_size_bits
    cdef uint64_t bit_mask

    cdef inline void _add(self, const uint64_t item)
    cdef inline bint _possibly_contains(self, const uint64_t item)
    cpdef numpy.ndarray[numpy.npy_bool, ndim=1] possibly_contains_many(self, object relation, list columns)

    cpdef void add(self, const uint64_t item)
    cpdef bint possibly_contains(self, const uint64_t item)

cpdef BloomFilter create_bloom_filter(object relation, list columns)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Show Create Node

This is a SQL Query Execution Plan Node.
"""

import pyarrow

from opteryx.exceptions import DatasetNotFoundError
from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.models import QueryProperties

from . import BasePlanNode


class ShowCreateNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **parameters):
        BasePlanNode.__init__(self, properties=properties, **parameters)

        self.object_type = parameters.get("object_type")
        self.object_name = parameters.get("object_name")

    @property
    def name(self):  # pragma: no cover
        return "Show"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self, morsel: pyarrow.Table, **kwargs) -> pyarrow.Table:
        if self.object_type == "VIEW":
            from opteryx.planner.views import is_view
            from opteryx.planner.views import view_as_sql

            if is_view(self.object_name):
                view_sql = view_as_sql(self.object_name)
                buffer = [{self.object_name: view_sql}]
                table = pyarrow.Table.from_pylist(buffer)
                yield table
                return

            raise DatasetNotFoundError(dataset=self.object_name, connector="VIEW")

        raise UnsupportedSyntaxError("Invalid SHOW statement")



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

import numpy
cimport numpy
numpy.import_array()

cpdef numpy.ndarray list_long_arrow_op(numpy.ndarray arr, object key):
    """
    Fetch values from a list of dictionaries based on a specified key.

    Parameters:
        data: list
            A list of dictionaries where each dictionary represents a structured record.
        key: str
            The key whose corresponding value is to be fetched from each dictionary.

    Returns:
        numpy.ndarray: An array containing the values associated with the key in each dictionary
                     or None where the key does not exist.
    """
    # Determine the number of items in the input list
    cdef Py_ssize_t n = len(arr)
    # Prepare an object array to store the results
    cdef numpy.ndarray result = numpy.empty(n, dtype=object)

    cdef Py_ssize_t i
    # Iterate over the list of dictionaries
    for i in range(n):
        # Check if the key exists in the dictionary
        if key in arr[i]:
            result[i] = str(arr[i][key])
        else:
            # Assign None if the key does not exist
            result[i] = None

    return result



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

import platform
from libcpp.vector cimport vector

cdef extern from "simd_search.h":
    size_t neon_count(const char* data, size_t length, char target)
    size_t avx_count(const char* data, size_t length, char target)
    int neon_search(const char* data, size_t length, char target)
    int avx_search(const char* data, size_t length, char target)
    int neon_find_delimiter(const char* data, size_t length)
    int avx_find_delimiter(const char* data, size_t length)
    vector[size_t] neon_find_all(const char* data, size_t length, char target)
    vector[size_t] avx_find_all(const char* data, size_t length, char target)

cdef extern from "simd_string_ops.h":
    void simd_to_upper(char* data, size_t length)
    void simd_to_lower(char* data, size_t length)

# Architecture detection
cdef size_t (*simd_count)(const char*, size_t, char)
cdef int (*simd_search)(const char*, size_t, char)
cdef int (*simd_find_delimiter)(const char*, size_t)
cdef vector[size_t] (*simd_find_all)(const char*, size_t, char)

_arch = platform.machine().lower()
if _arch in ('arm64', 'aarch64'):
    simd_count = neon_count
    simd_search = neon_search
    simd_find_delimiter = neon_find_delimiter
    simd_find_all = neon_find_all
else:
    simd_count = avx_count
    simd_search = avx_search
    simd_find_delimiter = avx_find_delimiter
    simd_find_all = avx_find_all


# Expose simd_count to Python so higher-level code can reuse the optimized newline/count routine.
cpdef size_t count_instances(const unsigned char[::1] mv, char target=10):
    """
    Count occurrences of `target` byte in `buffer` using the SIMD-optimized implementation.

    Returns an integer count.
    """
    cdef const unsigned char* udata = &mv[0]
    cdef const char* data = <const char*>udata
    cdef size_t data_len = mv.shape[0]
    return simd_count(data, data_len, target)


# String case conversion functions
def to_upper(bytes data):
    """
    Convert ASCII characters in bytes to uppercase using SIMD.
    Non-ASCII bytes are left unchanged.

    Returns a new bytes object with uppercase characters.
    """
    cdef size_t length = len(data)
    cdef bytearray result = bytearray(data)
    cdef char* ptr = <char*> result
    simd_to_upper(ptr, length)
    return bytes(result)


def to_lower(bytes data):
    """
    Convert ASCII characters in bytes to lowercase using SIMD.
    Non-ASCII bytes are left unchanged.

    Returns a new bytes object with lowercase characters.
    """
    cdef size_t length = len(data)
    cdef bytearray result = bytearray(data)
    cdef char* ptr = <char*> result
    simd_to_lower(ptr, length)
    return bytes(result)


# Character search functions
def find_char(bytes data, int target):
    """
    Find the first occurrence of a character in bytes using SIMD.

    Returns the index of the first occurrence, or -1 if not found.
    """
    cdef const char* ptr = <const char*> (<char*> data)
    cdef size_t length = len(data)
    cdef char char_target = <char> target
    return simd_search(ptr, length, char_target)


def count_char(bytes data, int target):
    """
    Count occurrences of a character in bytes using SIMD.

    Returns the count of occurrences.
    """
    cdef const char* ptr = <const char*> (<char*> data)
    cdef size_t length = len(data)
    cdef char char_target = <char> target
    return simd_count(ptr, length, char_target)


def find_all_char(bytes data, int target):
    """
    Find all occurrences of a character in bytes using SIMD.

    Returns a list of indices where the character occurs.
    """
    cdef const char* ptr = <const char*> (<char*> data)
    cdef size_t length = len(data)
    cdef char char_target = <char> target
    cdef vector[size_t] positions = simd_find_all(ptr, length, char_target)
    return list(positions)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.


from platform import node
from typing import Callable

from opteryx.managers.expression import format_expression
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

_render_registry: dict[LogicalPlanStepType, Callable[["LogicalPlanNode"], str]] = {}


def register_render(step_type: LogicalPlanStepType):
    """
    Decorator to register a rendering function for a given LogicalPlanStepType
    """

    def wrapper(func: Callable[["LogicalPlanNode"], str]):
        _render_registry[step_type] = func
        return func

    return wrapper


@register_render(LogicalPlanStepType.Filter)
def render_filter(node: LogicalPlanNode) -> str:
    return f"FILTER ({format_expression(node.condition)})"


@register_render(LogicalPlanStepType.Aggregate)
def render_aggregate(node: LogicalPlanNode) -> str:
    response = "AGGREGATE ["
    for col in node.aggregates:
        if col.condition:
            response += (
                f"{format_expression(col)} FILTER (WHERE {format_expression(col.condition)})"
            )
        else:
            response += format_expression(col)
        response += ", "
    response = response.rstrip(", ") + "]"
    return response


@register_render(LogicalPlanStepType.AggregateAndGroup)
def render_aggregate_group(node: LogicalPlanNode) -> str:
    aggregates = ", ".join(format_expression(col) for col in node.aggregates)
    groups = ", ".join(format_expression(col) for col in node.groups)
    return f"AGGREGATE [{aggregates}] GROUP BY [{groups}]"


@register_render(LogicalPlanStepType.Distinct)
def render_distinct(node: LogicalPlanNode) -> str:
    if node.on:
        cols = ",".join(format_expression(col) for col in node.on)
        return f"DISTINCT ON [{cols}]"
    return "DISTINCT"


@register_render(LogicalPlanStepType.Project)
def render_project(node: LogicalPlanNode) -> str:
    cols = ", ".join(format_expression(col) for col in node.columns)
    order_by = (
        f" + ({', '.join(format_expression(col) for col in node.order_by_columns)})"
        if node.order_by_columns
        else ""
    )
    except_cols = (
        f" EXCEPT ({', '.join(format_expression(col) for col in node.except_columns)})"
        if node.except_columns
        else ""
    )
    return f"PROJECT [{cols}]{except_cols}{order_by}"


@register_render(LogicalPlanStepType.Union)
def render_union(node: LogicalPlanNode) -> str:
    modifier = f" {node.modifier.upper()}" if node.modifier else ""
    columns = " [" + ", ".join(c.current_name for c in node.columns) + "]" if node.columns else ""
    return f"UNION{modifier}{columns}"


@register_render(LogicalPlanStepType.Explain)
def render_explain(node: LogicalPlanNode) -> str:
    fmt = f" (FORMAT {node.format})" if node.format else ""
    return f"EXPLAIN{' ANALYZE' if node.analyze else ''}{fmt}"


@register_render(LogicalPlanStepType.Difference)
def render_difference(_: LogicalPlanNode) -> str:
    return "DIFFERENCE"


@register_render(LogicalPlanStepType.Join)
def render_join(node: LogicalPlanNode) -> str:
    join_type = node.type.upper()
    cols = ""
    if node.columns:
        cols = ", ".join(format_expression(col) for col in node.columns)
        cols = f" [{cols}]"
    if node.on:
        return f"{join_type} JOIN ({format_expression(node.on, True)}){cols}"
    if node.using:
        using = ",".join(map(format_expression, node.using))
        return f"{join_type} JOIN (USING {using}){cols}"
    return f"{join_type} JOIN{cols}"


@register_render(LogicalPlanStepType.Unnest)
def render_unnest(node: LogicalPlanNode) -> str:
    distinct = "DISTINCT " if node.distinct else ""
    filters = f" FILTER ({', '.join(node.filters)})" if node.filters else ""
    return f"CROSS JOIN UNNEST ({distinct}{node.unnest_column.current_name}) AS {node.unnest_alias}{filters}"


@register_render(LogicalPlanStepType.AggregateAndGroup)
def render_aggregate_and_group(node: LogicalPlanNode) -> str:
    return f"AGGREGATE [{', '.join(format_expression(col) for col in node.aggregates)}] GROUP BY [{', '.join(format_expression(col) for col in node.groups)}]"


@register_render(LogicalPlanStepType.FunctionDataset)
def render_function_dataset(node: LogicalPlanNode) -> str:
    alias = f" AS {node.alias}" if node.alias else ""
    if node.function == "FAKE":
        return f"FAKE ({', '.join(format_expression(arg) for arg in node.args)}{alias})"
    if node.function == "GENERATE_SERIES":
        return f"GENERATE SERIES ({', '.join(format_expression(arg) for arg in node.args)}){alias}"
    if node.function == "VALUES":
        return f"VALUES (({', '.join(c.value for c in node.columns)}) x {len(node.values)} AS {node.alias})"
    if node.function == "UNNEST":
        return f"UNNEST ({', '.join(format_expression(arg) for arg in node.args)}{alias})"
    if node.function == "HTTP":
        return f"HTTP ({node.url}) AS {node.alias}"
    return node.function


@register_render(LogicalPlanStepType.HeapSort)
def render_heapsort(node: LogicalPlanNode) -> str:
    order = ", ".join(
        format_expression(expr) + (" DESC" if direction == "descending" else "")
        for expr, direction in node.order_by
    )
    return f"HEAP SORT (LIMIT {node.limit}, ORDER BY [{order}])"


@register_render(LogicalPlanStepType.Limit)
def render_limit(node: LogicalPlanNode) -> str:
    limit_str = f"LIMIT ({node.limit})" if node.limit is not None else ""
    offset_str = f" OFFSET ({node.offset})" if node.offset is not None else ""
    return (limit_str + offset_str).strip()


@register_render(LogicalPlanStepType.Order)
def render_order(node: LogicalPlanNode) -> str:
    order = ", ".join(
        format_expression(expr) + (" DESC" if direction == "descending" else "")
        for expr, direction in node.order_by
    )
    return f"ORDER BY [{order}]"


@register_render(LogicalPlanStepType.Scan)
def render_scan(node: LogicalPlanNode) -> str:
    io_async = "ASYNC " if hasattr(node.connector, "async_read_blob") else ""
    date_range = ""
    if node.start_date == node.end_date and node.start_date is not None:
        date_range = f" FOR '{node.start_date}'"
    elif node.start_date is not None:
        date_range = f" FOR '{node.start_date}' TO '{node.end_date}'"
    alias = f" AS {node.alias}" if node.relation != node.alias else ""
    columns = " [" + ", ".join(c.source_column for c in node.columns) + "]" if node.columns else ""
    predicates = (
        " (" + " AND ".join(map(format_expression, node.predicates)) + ")"
        if node.predicates
        else ""
    )
    hints = f" WITH({','.join(node.hints)})" if node.hints else ""
    limit = f" LIMIT {node.limit}" if node.limit else ""
    return f"{io_async}READ ({node.relation}{alias}{date_range}{hints}){columns}{predicates}{limit}"


@register_render(LogicalPlanStepType.Set)
def render_set(node: LogicalPlanNode) -> str:
    return f"SET ({node.variable} TO {node.value.value})"


@register_render(LogicalPlanStepType.Show)
def render_show(node: LogicalPlanNode) -> str:
    if node.object_type == "VARIABLE":
        return f"SHOW ({' '.join(node.items)})"
    if node.object_type == "VIEW":
        return f"SHOW (CREATE VIEW {node.object_name})"
    return "SHOW"


@register_render(LogicalPlanStepType.ShowColumns)
def render_show_columns(node: LogicalPlanNode) -> str:
    full = " FULL" if node.full else ""
    extended = " EXTENDED" if node.extended else ""
    return f"SHOW{full}{extended} COLUMNS ({node.relation})"


@register_render(LogicalPlanStepType.Subquery)
def render_subquery(node: LogicalPlanNode) -> str:
    return f"SUBQUERY{' AS ' + node.alias if node.alias else ''}"


@register_render(LogicalPlanStepType.CTE)
def render_cte(_: LogicalPlanNode) -> str:
    return "CTE"


@register_render(LogicalPlanStepType.MetadataWriter)
def render_metadata_writer(_: LogicalPlanNode) -> str:
    return "WRITE METADATA"


@register_render(LogicalPlanStepType.Exit)
def render_exit(_: LogicalPlanNode) -> str:
    return "EXIT"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Inner Join Node

This is a SQL Query Execution Plan Node.

This is an implementation of a hash join for the INNER JOIN operation in Opteryx. This is a
native implementation that does not use PyArrow's join capabilities. We heavily rely on
Cython to ensure performance.

This implementation includes the use of a bloom filter to quickly eliminate rows.

This implementation doesn't suffer from the limitations of PyArrow's join, such as the
inability to join on STRUCT or ARRAY columns.
"""

import time
from threading import Lock

import pyarrow
from pyarrow import Table

from opteryx import EOS
from opteryx.compiled.joins import build_side_hash_map
from opteryx.compiled.joins import get_last_inner_join_metrics
from opteryx.compiled.joins import inner_join
from opteryx.compiled.structures.bloom_filter import create_bloom_filter
from opteryx.models import QueryProperties
from opteryx.utils.arrow import align_tables

from . import JoinNode


class InnerJoinNode(JoinNode):
    join_type = "inner"

    def __init__(self, properties: QueryProperties, **parameters):
        JoinNode.__init__(self, properties=properties, **parameters)

        self.left_columns = parameters.get("left_columns")
        self.left_relation = None

        self.right_columns = parameters.get("right_columns")

        self.left_buffer = []
        self.left_buffer_columns = None
        self.left_hash = None
        self.left_filter = None

        self.columns = parameters.get("columns")

        self.lock = Lock()

    @property
    def name(self):  # pragma: no cover
        return "Inner Join"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self, morsel: Table, join_leg: str) -> Table:
        morsel = self.ensure_arrow_table(morsel)

        with self.lock:
            if join_leg == "left":
                if morsel == EOS:
                    self.left_relation = pyarrow.concat_tables(
                        self.left_buffer, promote_options="none"
                    )
                    self.left_buffer.clear()

                    start = time.monotonic_ns()
                    self.left_hash = build_side_hash_map(self.left_relation, self.left_columns)
                    self.statistics.increase(
                        "time_inner_join_build_side_hash_map",
                        time.monotonic_ns() - start,
                    )

                    # If the left side is small enough to quickly build a bloom filter, do that.
                    # - We use 16m + 1 as the upper limit to catch LIMIT on 16m rows
                    if self.left_relation.num_rows < 16_000_001:
                        start = time.monotonic_ns()
                        self.left_filter = create_bloom_filter(
                            self.left_relation, self.left_columns
                        )
                        self.statistics.increase(
                            "time_build_bloom_filter", time.monotonic_ns() - start
                        )
                        self.statistics.increase("feature_bloom_filter", 1)

                    # Project the left relation down to only the columns we need in the
                    # resulting morsel. This reduces the amount of data we keep in memory
                    # and pass through the align steps.
                    if self.columns is not None:
                        candidates = [c.schema_column.identity for c in self.columns]
                        left_keep = [c for c in candidates if c in self.left_relation.schema.names]
                        if len(left_keep) < len(self.left_relation.schema.names):
                            self.left_relation = self.left_relation.select(left_keep)
                            self.statistics.feature_eliminate_left_join_columns = 1
                else:
                    if self.left_buffer_columns is None:
                        self.left_buffer_columns = morsel.schema.names
                    else:
                        morsel = morsel.select(self.left_buffer_columns)
                    self.left_buffer.append(morsel)
                yield None
                return

            if join_leg == "right":
                if morsel == EOS:
                    yield EOS
                    return

                if self.left_filter is not None:
                    # Filter the morsel using the bloom filter, it's a quick way to
                    # reduce the number of rows that need to be joined.
                    start = time.monotonic_ns()
                    maybe_in_left = self.left_filter.possibly_contains_many(
                        morsel, self.right_columns
                    )
                    self.statistics.increase("time_bloom_filtering", time.monotonic_ns() - start)
                    morsel = morsel.filter(maybe_in_left)

                    # If the bloom filter is not effective, disable it.
                    # In basic benchmarks, the bloom filter is ~20x the speed of the join.
                    # so the break-even point is about 5% of the rows being eliminated.
                    eliminated_rows = len(maybe_in_left) - morsel.num_rows
                    if eliminated_rows < 0.05 * len(maybe_in_left):
                        self.left_filter = None
                        self.statistics.increase("feature_dynamically_disabled_bloom_filter", 1)

                    self.statistics.increase("rows_eliminated_by_bloom_filter", eliminated_rows)

                # do the join
                left_indicies, right_indicies = inner_join(
                    morsel, self.right_columns, self.left_hash
                )

                # record detailed timing and row counts for diagnostics
                (
                    hash_time,
                    probe_time,
                    rows_hashed,
                    candidate_rows,
                    matched_rows,
                    materialize_time,
                ) = get_last_inner_join_metrics()
                self.statistics.increase("time_inner_join_hash", hash_time)
                self.statistics.increase("time_inner_join_probe", probe_time)
                self.statistics.increase("rows_inner_join_hashed", rows_hashed)
                self.statistics.increase("rows_inner_join_candidates", candidate_rows)
                self.statistics.increase("time_inner_join_indices", materialize_time)
                self.statistics.increase("rows_inner_join_matched", matched_rows)
                start = time.monotonic_ns()

                # Project the right relation down to only the columns we need in the
                # resulting morsel. This reduces the amount of data we keep in memory
                # and pass through the align steps.
                if self.columns is not None:
                    candidates = [c.schema_column.identity for c in self.columns]
                    right_keep = [c for c in candidates if c in morsel.schema.names]
                    if len(right_keep) < len(morsel.schema.names):
                        morsel = morsel.select(right_keep)
                        self.statistics.feature_eliminate_right_join_columns = 1

                aligned = align_tables(morsel, self.left_relation, right_indicies, left_indicies)
                self.statistics.increase("time_inner_join_align", time.monotonic_ns() - start)

                yield aligned



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Iceberg Connector
"""

import datetime
import struct
from decimal import Decimal
from typing import Dict
from typing import Union

import numpy
import pyarrow
from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.tools import single_item_cache
from orso.types import OrsoTypes

from opteryx.connectors.base.base_connector import BaseConnector
from opteryx.connectors.capabilities import Diachronic
from opteryx.connectors.capabilities import LimitPushable
from opteryx.connectors.capabilities import PredicatePushable
from opteryx.connectors.capabilities import Statistics
from opteryx.exceptions import DatasetNotFoundError
from opteryx.exceptions import DatasetReadError
from opteryx.exceptions import NotSupportedError
from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import RelationStatistics
from opteryx.utils.file_decoders import filter_records


@single_item_cache
def to_iceberg_filter(root):
    """
    Convert a filter to Iceberg filter form.

    This is specifically opinionated for the Iceberg reader.
    """
    import pyiceberg
    import pyiceberg.expressions

    ICEBERG_FILTERS = {
        "GtEq": pyiceberg.expressions.GreaterThanOrEqual,
        "Eq": pyiceberg.expressions.EqualTo,
        "Gt": pyiceberg.expressions.GreaterThan,
        "Lt": pyiceberg.expressions.LessThan,
        "LtEq": pyiceberg.expressions.LessThanOrEqual,
        "NotEq": pyiceberg.expressions.NotEqualTo,
    }

    def _predicate_to_iceberg_filter(root):
        # Reduce look-ahead effort by using Exceptions to control flow
        if root.node_type == NodeType.AND:  # pragma: no cover
            left = _predicate_to_iceberg_filter(root.left)
            right = _predicate_to_iceberg_filter(root.right)
            if not isinstance(left, list):
                left = [left]
            if not isinstance(right, list):
                right = [right]
            left.extend(right)
            return left
        if root.node_type != NodeType.COMPARISON_OPERATOR:
            raise NotSupportedError()
        if root.left.node_type != NodeType.IDENTIFIER:
            root.left, root.right = root.right, root.left
        if root.right.schema_column.type == OrsoTypes.DATE:
            date_val = root.right.value
            if hasattr(date_val, "item"):
                date_val = date_val.item()
            root.right.value = datetime.datetime.combine(date_val, datetime.time.min)
            root.right.schema_column.type = OrsoTypes.TIMESTAMP
        if root.left.schema_column.type == OrsoTypes.DATE:
            root.left.schema_column.type = OrsoTypes.TIMESTAMP
        if root.left.node_type != NodeType.IDENTIFIER:
            raise NotSupportedError()
        if root.right.node_type != NodeType.LITERAL:
            raise NotSupportedError()
        if root.left.schema_column.type == OrsoTypes.VARCHAR:
            root.left.schema_column.type = OrsoTypes.BLOB
        if root.right.schema_column.type == OrsoTypes.VARCHAR:
            root.right.schema_column.type = OrsoTypes.BLOB
        if root.right.schema_column.type != root.left.schema_column.type:
            raise NotSupportedError(
                f"{root.right.schema_column.type} != {root.left.schema_column.type}"
            )
        if root.right.schema_column.type == OrsoTypes.DOUBLE:
            # iceberg needs doubles to be cast to floats
            root.right.value = float(root.right.value)
        if root.right.schema_column.type == OrsoTypes.INTEGER:
            # iceberg doesn't like integers unless we convert to strings
            root.right.value = str(root.right.value)
        if root.right.schema_column.type == OrsoTypes.TIMESTAMP:
            # iceberg doesn't like timestamps unless we convert to strings
            if isinstance(root.right.value, numpy.datetime64):
                root.right.value = root.right.value.astype(datetime.datetime)
            root.right.value = root.right.value.isoformat()
        return ICEBERG_FILTERS[root.value](root.left.value, root.right.value)

    iceberg_filter = None
    unsupported = []
    if not isinstance(root, list):
        root = [root]
    for predicate in root:
        try:
            converted = _predicate_to_iceberg_filter(predicate)
            if iceberg_filter is None:
                iceberg_filter = converted
            else:
                iceberg_filter = pyiceberg.expressions.And(iceberg_filter, converted)
        except NotSupportedError:
            unsupported.append(predicate)

    return iceberg_filter if iceberg_filter else "True", unsupported


class IcebergConnector(BaseConnector, Diachronic, LimitPushable, Statistics, PredicatePushable):
    __mode__ = "Blob"
    __type__ = "ICEBERG"

    PUSHABLE_OPS: Dict[str, bool] = {
        "Eq": True,
        "NotEq": True,  # nulls not handled correctly
        "Gt": True,
        "GtEq": True,
        "Lt": True,
        "LtEq": True,
    }

    PUSHABLE_TYPES = {
        OrsoTypes.BLOB,
        OrsoTypes.BOOLEAN,
        OrsoTypes.DOUBLE,
        OrsoTypes.INTEGER,
        OrsoTypes.VARCHAR,
        OrsoTypes.TIMESTAMP,
        OrsoTypes.DATE,
    }

    def __init__(self, *args, catalog=None, **kwargs):
        BaseConnector.__init__(self, **kwargs)
        LimitPushable.__init__(self, **kwargs)
        Diachronic.__init__(self, **kwargs)
        Statistics.__init__(self, **kwargs)
        PredicatePushable.__init__(self, **kwargs)

        import pyiceberg

        try:
            self.table = catalog.load_table(self.dataset)
            self.snapshot = self.table.current_snapshot()
            # If the table exists but has no snapshots, we don't raise here —
            # we allow non-time-travel reads to return an empty result set with
            # a valid schema. For time-travel (start_date specified), we will
            # still raise DatasetReadError in get_dataset_schema.
            self.snapshot_id = None if self.snapshot is None else self.snapshot.snapshot_id
        except pyiceberg.exceptions.NoSuchTableError:
            raise DatasetNotFoundError(dataset=self.dataset, connector=self.__type__) from None

    def get_dataset_schema(self) -> RelationSchema:
        if self.start_date != self.end_date:
            if self.start_date.date() != self.end_date.date():
                raise UnsupportedSyntaxError("This table only supports point in time reads.")
            raise UnsupportedSyntaxError(
                "This table only supports point in time reads. Are you missing the time component from your FOR clause?"
            )

        if self.start_date is not None:
            snapshots = self.table.inspect.snapshots().sort_by("committed_at")
            snapshot_rows = snapshots.to_pylist()

            if not snapshot_rows:
                raise DatasetReadError("No data available for the specified date.")

            # Honor dates before the first snapshot by rejecting them, but treat
            # dates after the latest snapshot as selecting the latest snapshot
            first_committed = snapshot_rows[0]["committed_at"]
            last_committed = snapshot_rows[-1]["committed_at"]

            if self.start_date < first_committed:
                # Point-in-time read is before our first snapshot — no data available then
                raise DatasetReadError("No data available for the specified date.")
            elif self.start_date > last_committed:
                # Point-in-time read after the latest snapshot — return current data
                selected = snapshot_rows[-1]
                # ensure we store the commit time for statistics/context
                self.statistics.dataset_committed_at = selected["committed_at"].isoformat()
            else:
                selected = snapshot_rows[0]
                for candidate in snapshot_rows:
                    if candidate["committed_at"] <= self.start_date:
                        self.statistics.dataset_committed_at = candidate["committed_at"].isoformat()
                        selected = candidate
                    else:
                        break

            self.snapshot_id = selected["snapshot_id"]
            self.snapshot = self.table.snapshot_by_id(self.snapshot_id)

        # If the table has no snapshot and the read is not time-travel, use
        # the table's declared schema (from metadata) and return an empty result set.
        if self.snapshot is None:
            iceberg_schema = self.table.schema()
        else:
            iceberg_schema = self.table.schemas()[self.snapshot.schema_id]
            self.statistics.dataset_committed_at = datetime.datetime.fromtimestamp(
                self.snapshot.timestamp_ms / 1000.0
            ).isoformat()
        arrow_schema = iceberg_schema.as_arrow()

        self.schema = RelationSchema(
            name=self.dataset,
            columns=[FlatColumn.from_arrow(field) for field in arrow_schema],
        )

        # Get statistics
        relation_statistics = RelationStatistics()

        column_names = {col.field_id: col.name for col in iceberg_schema.columns}
        column_types = {col.field_id: col.field_type for col in iceberg_schema.columns}

        files = self.table.inspect.files(snapshot_id=self.snapshot_id)

        # No files = empty table, no stats
        if len(files.column("file_path")) == 0:
            self.relation_statistics = relation_statistics
            return self.schema

        relation_statistics.record_count = pyarrow.compute.sum(files.column("record_count")).as_py()

        if "distinct_counts" in files.columns:
            for file in files.column("distinct_counts"):
                for k, v in file:
                    relation_statistics.set_cardinality_estimate(column_names[k], v)

        if "value_counts" in files.columns:
            for file in files.column("value_counts"):
                for k, v in file:
                    relation_statistics.add_count(column_names[k], v)

        for file in files.column("lower_bounds"):
            for k, v in file:
                relation_statistics.update_lower(
                    column_names[k], IcebergConnector.decode_iceberg_value(v, column_types[k])
                )

        for file in files.column("upper_bounds"):
            for k, v in file:
                relation_statistics.update_upper(
                    column_names[k], IcebergConnector.decode_iceberg_value(v, column_types[k])
                )

        self.relation_statistics = relation_statistics

        return self.schema

    def read_dataset(
        self, columns: list = None, predicates: list = None, limit: int = None, **kwargs
    ) -> pyarrow.Table:
        rows_read = 0

        if columns is None:
            column_names = self.schema.column_names
        else:
            column_names = [col.source_column for col in columns]

        pushed_filters, unsupported = to_iceberg_filter(predicates)
        selected_columns = list(
            set(column_names).union(
                {
                    c.source_column
                    for c in get_all_nodes_of_type(unsupported, (NodeType.IDENTIFIER,))
                }
            )
        )

        # Short-cut COUNT(*) handling
        if selected_columns == []:
            table = pyarrow.Table.from_arrays(
                [[self.relation_statistics.record_count]], names=["$COUNT(*)"]
            )
            yield table
            return

        reader = self.table.scan(
            row_filter=pushed_filters,
            selected_fields=selected_columns,
            limit=limit,
            snapshot_id=self.snapshot_id,
        ).to_arrow_batch_reader()

        # If there are no snapshots (snapshot_id is None), return an empty morsel
        if self.snapshot_id is None:
            from orso.schema import RelationSchema
            from orso.schema import convert_orso_schema_to_arrow_schema

            orso_schema = RelationSchema(
                name="Relation", columns=[c.schema_column for c in columns]
            )
            arrow_shema = convert_orso_schema_to_arrow_schema(orso_schema, use_identities=True)

            morsel = pyarrow.Table.from_arrays(
                [pyarrow.array([]) for _ in columns],
                schema=arrow_shema,
            )
            yield morsel
            return

        batch = None
        for batch in reader:
            # Check for decimal columns in the batch schema
            for field in batch.schema:
                if pyarrow.types.is_decimal(field.type):
                    raise NotSupportedError("Decimal columns are not supported in Iceberg tables.")

            table = pyarrow.Table.from_batches([batch])

            if unsupported:
                table = filter_records(unsupported, table)
            yield table
            rows_read += batch.num_rows

        if batch is None:
            from orso.schema import RelationSchema
            from orso.schema import convert_orso_schema_to_arrow_schema

            orso_schema = RelationSchema(
                name="Relation", columns=[c.schema_column for c in columns]
            )
            arrow_shema = convert_orso_schema_to_arrow_schema(orso_schema, use_identities=True)

            morsel = pyarrow.Table.from_arrays(
                [pyarrow.array([]) for _ in columns],
                schema=arrow_shema,
            )
            yield morsel

    @staticmethod
    def decode_iceberg_value(
        value: Union[int, float, bytes], data_type: str, scale: int = None
    ) -> Union[int, float, str, datetime.datetime, Decimal, bool]:
        """
        Decode Iceberg-encoded values based on the specified data type.

        Parameters:
            value: Union[int, float, bytes]
                The encoded value from Iceberg.
            data_type: str
                The type of the value ('int', 'long', 'float', 'double', 'timestamp', 'date', 'string', 'decimal', 'boolean').
            scale: int, optional
                Scale used for decoding decimal types, defaults to None.

        Returns:
            The decoded value in its original form.
        """
        import pyiceberg

        data_type_class = data_type.__class__

        if data_type_class == pyiceberg.types.LongType:
            return int.from_bytes(value, "little", signed=True)
        elif data_type_class == pyiceberg.types.DoubleType:
            # IEEE 754 encoded floats are typically decoded directly
            return struct.unpack("<d", value)[0]  # 8-byte IEEE 754 double
        elif data_type_class in (pyiceberg.types.TimestampType, pyiceberg.types.TimestamptzType):
            # Iceberg stores timestamps as microseconds since epoch
            interval = int.from_bytes(value, "little", signed=True)
            if interval < 0:
                # Windows specifically doesn't like negative timestamps
                return datetime.datetime(1970, 1, 1) + datetime.timedelta(microseconds=interval)
            return datetime.datetime.fromtimestamp(interval / 1_000_000)
        elif data_type == "date":
            # Iceberg stores dates as days since epoch (1970-01-01)
            interval = int.from_bytes(value, "little", signed=True)
            return datetime.datetime(1970, 1, 1) + datetime.timedelta(days=interval)
        elif data_type_class == pyiceberg.types.StringType:
            # Assuming UTF-8 encoded bytes (or already decoded string)
            return value.decode("utf-8") if isinstance(value, bytes) else str(value)
        elif data_type_class == pyiceberg.types.BinaryType:
            return value
        elif str(data_type).startswith("decimal"):
            # Iceberg stores decimals as unscaled integers
            int_value = int.from_bytes(value, byteorder="big", signed=True)
            return Decimal(int_value) / (10**data_type.scale)
        elif data_type_class == pyiceberg.types.BooleanType:
            return bool(value)

        ValueError(f"Unsupported data type: {data_type}, {str(data_type)}")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Compatibility layer exposing the existing Opteryx connection and cursor objects
under the conventional DBAPI module namespace.
"""

from typing import Any

from orso.types import OrsoTypes

import opteryx
from opteryx import connect
from opteryx import exceptions as exc
from opteryx.connection import Connection
from opteryx.cursor import Cursor

apilevel = opteryx.apilevel
threadsafety = opteryx.threadsafety
paramstyle = opteryx.paramstyle


class DBAPISet(frozenset):
    """PEP-249 helper to compare database type categories."""

    def __eq__(self, other: object) -> bool:
        if isinstance(other, frozenset):
            return super().__eq__(other)
        return other in self

    def __ne__(self, other: object) -> bool:
        if isinstance(other, set):
            return super().__ne__(other)
        return other not in self


STRING = DBAPISet([OrsoTypes.VARCHAR, OrsoTypes.JSONB])
BINARY = DBAPISet([OrsoTypes.BLOB])
NUMBER = DBAPISet([OrsoTypes.INTEGER, OrsoTypes.DOUBLE, OrsoTypes.DECIMAL])
DATE = DBAPISet([OrsoTypes.DATE])
TIME = DBAPISet([OrsoTypes.TIME])
TIMESTAMP = DBAPISet([OrsoTypes.TIMESTAMP])
DATETIME = TIMESTAMP
ROWID = DBAPISet()


def Binary(value: Any) -> bytes:
    """Return ``value`` as a binary object, per DBAPI 2.0."""

    return bytes(value)


Warning = exc.SecurityError  # type: ignore
Error = exc.Error
InterfaceError = exc.ProgrammingError
DatabaseError = exc.DatabaseError
OperationalError = exc.ExecutionError
IntegrityError = exc.DataError
InternalError = exc.InvalidInternalStateError  # type: ignore
DataError = exc.DataError
ProgrammingError = exc.ProgrammingError
NotSupportedError = exc.NotSupportedError

__all__ = [
    "apilevel",
    "threadsafety",
    "paramstyle",
    "Connection",
    "Cursor",
    "connect",
    "Binary",
    "STRING",
    "BINARY",
    "NUMBER",
    "DATE",
    "TIME",
    "TIMESTAMP",
    "DATETIME",
    "ROWID",
    "Warning",
    "Error",
    "InterfaceError",
    "DatabaseError",
    "OperationalError",
    "IntegrityError",
    "InternalError",
    "DataError",
    "ProgrammingError",
    "NotSupportedError",
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Stop Words - words not included in the VECTOR INDEX or in MATCH AGAINST searches
"""

from typing import Set

from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.types import OrsoTypes

from opteryx.models import RelationStatistics

__all__ = ("read", "schema")

STOP_WORDS: Set[bytes] = {
    b"about",
    b"above",
    b"across",
    b"after",
    b"afterwards",
    b"again",
    b"against",
    b"all",
    b"almost",
    b"alone",
    b"along",
    b"already",
    b"also",
    b"although",
    b"always",
    b"am",
    b"among",
    b"amongst",
    b"amount",
    b"an",
    b"and",
    b"another",
    b"any",
    b"anyhow",
    b"anyone",
    b"anything",
    b"anyway",
    b"anywhere",
    b"are",
    b"around",
    b"as",
    b"at",
    b"back",
    b"be",
    b"became",
    b"because",
    b"become",
    b"becomes",
    b"becoming",
    b"been",
    b"before",
    b"beforehand",
    b"behind",
    b"being",
    b"below",
    b"beside",
    b"besides",
    b"between",
    b"beyond",
    b"both",
    b"bottom",
    b"but",
    b"by",
    b"ca",
    b"call",
    b"can",
    b"cannot",
    b"could",
    b"did",
    b"do",
    b"does",
    b"doing",
    b"done",
    b"down",
    b"due",
    b"during",
    b"each",
    b"eight",
    b"either",
    b"eleven",
    b"else",
    b"elsewhere",
    b"empty",
    b"enough",
    b"even",
    b"ever",
    b"every",
    b"everyone",
    b"everything",
    b"everywhere",
    b"except",
    b"few",
    b"fifteen",
    b"fifty",
    b"first",
    b"five",
    b"for",
    b"former",
    b"formerly",
    b"forty",
    b"four",
    b"from",
    b"front",
    b"full",
    b"further",
    b"get",
    b"give",
    b"go",
    b"had",
    b"has",
    b"have",
    b"he",
    b"hence",
    b"her",
    b"here",
    b"hereafter",
    b"hereby",
    b"herein",
    b"hereupon",
    b"hers",
    b"herself",
    b"him",
    b"himself",
    b"his",
    b"how",
    b"however",
    b"hundred",
    b"if",
    b"in",
    b"indeed",
    b"into",
    b"is",
    b"it",
    b"its",
    b"itself",
    b"just",
    b"keep",
    b"last",
    b"latter",
    b"latterly",
    b"least",
    b"less",
    b"ll",
    b"made",
    b"make",
    b"many",
    b"may",
    b"me",
    b"meanwhile",
    b"might",
    b"mine",
    b"more",
    b"moreover",
    b"most",
    b"mostly",
    b"move",
    b"much",
    b"must",
    b"my",
    b"myself",
    b"name",
    b"namely",
    b"neither",
    b"never",
    b"nevertheless",
    b"next",
    b"nine",
    b"no",
    b"nobody",
    b"none",
    b"noone",
    b"nor",
    b"not",
    b"nothing",
    b"now",
    b"nowhere",
    b"of",
    b"off",
    b"often",
    b"on",
    b"once",
    b"one",
    b"only",
    b"onto",
    b"or",
    b"other",
    b"others",
    b"otherwise",
    b"our",
    b"ours",
    b"ourselves",
    b"out",
    b"over",
    b"own",
    b"part",
    b"per",
    b"perhaps",
    b"please",
    b"put",
    b"quite",
    b"rather",
    b"re",
    b"really",
    b"regarding",
    b"same",
    b"say",
    b"see",
    b"seem",
    b"seemed",
    b"seeming",
    b"seems",
    b"serious",
    b"several",
    b"she",
    b"should",
    b"show",
    b"side",
    b"since",
    b"six",
    b"sixty",
    b"so",
    b"some",
    b"somehow",
    b"someone",
    b"something",
    b"sometime",
    b"sometimes",
    b"somewhere",
    b"still",
    b"such",
    b"take",
    b"ten",
    b"than",
    b"that",
    b"the",
    b"their",
    b"them",
    b"themselves",
    b"then",
    b"thence",
    b"there",
    b"thereafter",
    b"thereby",
    b"therefore",
    b"therein",
    b"thereupon",
    b"these",
    b"they",
    b"third",
    b"this",
    b"those",
    b"though",
    b"three",
    b"through",
    b"throughout",
    b"thru",
    b"thus",
    b"to",
    b"together",
    b"too",
    b"top",
    b"toward",
    b"towards",
    b"twelve",
    b"twenty",
    b"two",
    b"under",
    b"unless",
    b"until",
    b"up",
    b"upon",
    b"us",
    b"used",
    b"using",
    b"various",
    b"very",
    b"via",
    b"ve",
    b"was",
    b"we",
    b"well",
    b"were",
    b"what",
    b"whatever",
    b"when",
    b"whence",
    b"whenever",
    b"where",
    b"whereafter",
    b"whereas",
    b"whereby",
    b"wherein",
    b"whereupon",
    b"wherever",
    b"whether",
    b"which",
    b"while",
    b"whither",
    b"who",
    b"whoever",
    b"whole",
    b"whom",
    b"whose",
    b"why",
    b"will",
    b"with",
    b"within",
    b"without",
    b"would",
    b"yet",
    b"you",
    b"your",
    b"yours",
    b"yourself",
    b"yourselves",
}


def read(end_date=None, *args):
    import pyarrow

    # Define the data
    data = [
        pyarrow.array(
            STOP_WORDS,
            type=pyarrow.string(),
        ),
    ]
    column_names = ["value"]

    return pyarrow.Table.from_arrays(data, column_names)


def schema():
    # fmt:off
    return RelationSchema(
            name="$stop_words",
            columns=[
                FlatColumn(name="value", type=OrsoTypes.VARCHAR),
            ],
        )
    # fmt:on


def statistics() -> RelationStatistics:
    return RelationStatistics()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
missions
----------

This is a sample dataset build into the engine, this simplifies a few things:

- We can write test scripts using this data, knowing that it will always be available.
- We can write examples using this data, knowing the results will always match.

Space Mission dataset acquired from [Kaggle](https://www.kaggle.com/datasets/agirlcoding/all-space-missions-from-1957).
"""

import datetime

from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.tools import single_item_cache
from orso.types import OrsoTypes

from opteryx.models import RelationStatistics

__all__ = ("read", "schema")

_decoded: bytes = None


@single_item_cache
def read(*args):
    from opteryx.virtual_datasets import load_virtual_dataset

    return load_virtual_dataset("$missions")


def schema():
    # fmt:off
    return RelationSchema(
            name="$missions",
            columns=[
                FlatColumn(name="Company", type=OrsoTypes.VARCHAR),
                FlatColumn(name="Location", type=OrsoTypes.VARCHAR),
                FlatColumn(name="Price", type=OrsoTypes.DOUBLE),
                FlatColumn(name="Lauched_at", type=OrsoTypes.TIMESTAMP, aliases=["Launched_at"]),
                FlatColumn(name="Rocket", type=OrsoTypes.VARCHAR),
                FlatColumn(name="Rocket_Status", type=OrsoTypes.VARCHAR),
                FlatColumn(name="Mission", type=OrsoTypes.VARCHAR),
                FlatColumn(name="Mission_Status", type=OrsoTypes.VARCHAR),
            ],
        )


def statistics() -> RelationStatistics:
    from opteryx.compiled.structures.relation_statistics import to_int

    stats = RelationStatistics()

    stats.record_count = 4630
    stats.lower_bounds = {
        b"Company": to_int("AEB"),
        b"Location": to_int("Blue Origin Launch Site, West Texas, Texas, USA"),
        b"Price": to_int(2.5),
        b"Lauched_at": to_int(datetime.datetime(1957, 10, 4, 19, 28)),
        b"Rocket": to_int("ASLV"),
        b"Rocket_Status": to_int("Active"),
        b"Mission": to_int("-TJS_6.00"),
        b"Mission_Status": to_int("Failure"),
    }
    stats.upper_bounds = {
        b"Company": to_int("i-Space"),
        b"Location": to_int("Xichang Satellite Launch Center, China"),
        b"Price": to_int(450.0),
        b"Lauched_at": to_int(datetime.datetime(2022, 7, 29, 13, 28)),
        b"Rocket": to_int("Zoljanah"),
        b"Rocket_Status": to_int("Retired"),
        b"Mission": to_int("iPStar-1"),
        b"Mission_Status": to_int("Success"),
    }
    stats.null_count = {
        b"Company": 0,
        b"Location": 0,
        b"Price": 3380,
        b"Lauched_at": 127,
        b"Rocket": 0,
        b"Rocket_Status": 0,
        b"Mission": 0,
        b"Mission_Status": 0,
    }

    return stats



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Union Node

This is a SQL Query Execution Plan Node.
"""

from pyarrow import Table

from opteryx import EOS
from opteryx.models import QueryProperties

from . import BasePlanNode


class UnionNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **parameters):
        BasePlanNode.__init__(self, properties=properties, **parameters)
        self.columns = parameters.get("columns", [])
        self.column_ids = [c.schema_column.identity for c in self.columns]
        self.seen_first_eos = False
        self.schema = None

    @property
    def name(self):  # pragma: no cover
        return "Union"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self, morsel: Table, **kwargs) -> Table:
        """
        Union needs to ensure the column names are the same and that
        coercible types are coerced.
        """
        morsel = self.ensure_arrow_table(morsel)

        if morsel == EOS and self.seen_first_eos:
            yield EOS
            return
        elif morsel == EOS:
            self.seen_first_eos = True
            yield None
            return

        elif self.schema is None:
            self.schema = morsel.schema
        else:
            morsel = morsel.rename_columns(self.schema.names)
            morsel = morsel.cast(self.schema)

        if len(morsel.schema.names) != len(self.column_ids):
            morsel = morsel.select(self.schema.names[: len(self.column_ids)])
        morsel = morsel.rename_columns(self.column_ids)
        yield morsel.select(self.column_ids)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Heap Sort Node (Top-N Sort)

This is a SQL Query Execution Plan Node.

This query plan node maintains a sorted table of the top-N records seen so far
based on the provided ORDER BY clause. Despite the name, this is not a Heap Sort
algorithm, but an incremental Top-N sorter that works chunk-wise for efficiency.

This is faster, particularly when working with large datasets even though we're now
sorting smaller chunks over and over again.
"""

import pyarrow
import pyarrow.compute as pc
from pyarrow import concat_tables

from opteryx import EOS
from opteryx.exceptions import ColumnNotFoundError
from opteryx.models import QueryProperties

from . import BasePlanNode


class HeapSortNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **parameters):
        super().__init__(properties=properties, **parameters)
        self.order_by = parameters.get("order_by", [])
        self.limit = parameters.get("limit", -1)

        self.mapped_order = []
        for column, direction in self.order_by:
            try:
                self.mapped_order.append((column.schema_column.identity, direction))
            except ColumnNotFoundError as cnfe:
                raise ColumnNotFoundError(
                    f"`ORDER BY` must reference columns from `SELECT`. {cnfe}"
                ) from cnfe
        self.table = None

    @property
    def config(self):  # pragma: no cover
        order = ", ".join(
            f"{col.schema_column.name} {dir[:3].upper()}" for col, dir in self.order_by
        )
        return f"LIMIT = {self.limit}, ORDER = {order}"

    @property
    def name(self):  # pragma: no cover
        return "Heap Sort"

    def execute(self, morsel: pyarrow.Table, **kwargs):
        morsel = self.ensure_arrow_table(morsel)

        _ = kwargs  # kwargs are part of the execution contract
        if morsel is EOS:
            if self.table is None:
                yield EOS
                return

            if (self.limit is None or self.limit <= 0) and self.mapped_order:
                self.table = self.table.sort_by(self.mapped_order)
            elif self.limit and self.limit > 0 and self.mapped_order:
                # Ensure sorted output for the final chunk
                self.table = self.table.sort_by(self.mapped_order)
                if self.table.num_rows > self.limit:
                    self.table = self.table.slice(0, self.limit)

            yield self.table
            yield EOS
            return

        if morsel.num_rows == 0:
            yield None
            return

        if self.limit and self.limit > 0:
            morsel = self._top_n(morsel)
            if self.table is None:
                self.table = morsel
            else:
                combined = concat_tables([self.table, morsel], promote_options="permissive")
                self.table = self._top_n(combined)
        else:
            if self.table is None:
                self.table = morsel
            else:
                self.table = concat_tables([self.table, morsel], promote_options="permissive")

        yield None

    def _top_n(self, table: pyarrow.Table) -> pyarrow.Table:
        if self.limit is None or self.limit <= 0:
            return table

        k = min(self.limit, table.num_rows)
        if k == 0:
            return table.slice(0, 0)

        if not self.mapped_order:
            return table.slice(0, k)

        if k < table.num_rows:
            try:
                indices = pc.select_k_unstable(  # type: ignore[attr-defined]
                    table,
                    k=k,
                    sort_keys=self.mapped_order,
                )
            except AttributeError:
                limited = table.sort_by(self.mapped_order).slice(0, k)
            except (TypeError, ValueError, NotImplementedError):
                limited = table.sort_by(self.mapped_order).slice(0, k)
            else:
                limited = table.take(indices)
                return limited.sort_by(self.mapped_order)

            return limited

        sorted_table = table.sort_by(self.mapped_order)
        return sorted_table.slice(0, k)



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

"""
BoolVector: Cython implementation of a zero-copy, bit-packed boolean column vector for Draken.

This matches Arrow's representation:
- Values are stored bit-packed in data buffer (1 bit per value).
- Nulls are stored in the null_bitmap (same layout).
- Zero-copy interop with Arrow via from_arrow/to_arrow.

"""

from cpython.bytes cimport PyBytes_FromStringAndSize, PyBytes_AS_STRING
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.string cimport memset

from libc.stdint cimport int32_t, int8_t, intptr_t, uint64_t, uint8_t
from libc.stdlib cimport malloc

from opteryx.draken.core.buffers cimport DrakenFixedBuffer
from opteryx.draken.core.buffers cimport DRAKEN_BOOL
from opteryx.draken.core.fixed_vector cimport alloc_fixed_buffer, buf_dtype, buf_length, free_fixed_buffer
from opteryx.draken.vectors.vector cimport MIX_HASH_CONSTANT, Vector, NULL_HASH, mix_hash, simd_mix_hash

cdef const uint64_t TRUE_HASH = <uint64_t>0x4f112caa54efa882
cdef const uint64_t FALSE_HASH = <uint64_t>0xc2fd8b2343f83ce7

DEF BOOL_HASH_CHUNK = 1024

cdef class BoolVector(Vector):

    def __cinit__(self, size_t length=0, bint wrap=False):
        cdef size_t nbytes

        if wrap:
            self.ptr = NULL
            self.owns_data = False
        else:
            # bit-packed, so allocate ceil(length/8) bytes
            nbytes = (length + 7) >> 3
            self.ptr = alloc_fixed_buffer(DRAKEN_BOOL, length, 1)  # itemsize=1 is logical
            if self.ptr != NULL:
                # allocate raw bytes with libc malloc so free_fixed_buffer (which calls free())
                # can safely free the buffer later. Do not mix Python allocator and free().
                self.ptr.data = malloc(nbytes)
                if self.ptr.data == NULL:
                    raise MemoryError()
            self.owns_data = True

    def __dealloc__(self):
        if self.owns_data and self.ptr is not NULL:
            free_fixed_buffer(self.ptr, True)
            self.ptr = NULL

    # Properties
    @property
    def length(self):
        return buf_length(self.ptr)

    @property
    def dtype(self):
        return buf_dtype(self.ptr)

    def __getitem__(self, Py_ssize_t i):
        """Return the value at index i, or None if null."""
        cdef DrakenFixedBuffer* ptr = self.ptr
        if i < 0 or i >= ptr.length:
            raise IndexError("Index out of bounds")
        # null check
        if ptr.null_bitmap != NULL:
            byte = ptr.null_bitmap[i >> 3]
            bit = (byte >> (i & 7)) & 1
            if not bit:
                return None
        # extract bit
        cdef uint8_t val_byte = (<uint8_t*>ptr.data)[i >> 3]
        return bool((val_byte >> (i & 7)) & 1)

    # -------- Interop --------
    def to_arrow(self):
        """Convert to a PyArrow array."""
        import pyarrow as pa
        
        cdef size_t nbytes = (buf_length(self.ptr) + 7) >> 3
        addr = <intptr_t> self.ptr.data
        data_buf = pa.foreign_buffer(addr, nbytes, base=self)

        buffers = []
        if self.ptr.null_bitmap != NULL:
            buffers.append(pa.foreign_buffer(<intptr_t> self.ptr.null_bitmap, (self.ptr.length + 7) // 8, base=self))
        else:
            buffers.append(None)

        buffers.append(data_buf)

        return pa.Array.from_buffers(pa.bool_(), buf_length(self.ptr), buffers)

    cpdef BoolVector and_vector(self, BoolVector other):
        """Element-wise AND between two BoolVector instances. Returns a new BoolVector."""
        cdef DrakenFixedBuffer* ptr1 = self.ptr
        cdef DrakenFixedBuffer* ptr2 = other.ptr
        cdef Py_ssize_t n = ptr1.length
        if n != ptr2.length:
            raise ValueError("Vectors must have the same length")

        cdef Py_ssize_t nbytes = (n + 7) >> 3
        cdef BoolVector out = BoolVector(<size_t>n)
        cdef uint8_t* a = <uint8_t*> ptr1.data
        cdef uint8_t* b = <uint8_t*> ptr2.data
        cdef uint8_t* d = <uint8_t*> out.ptr.data
        cdef Py_ssize_t i
        for i in range(nbytes):
            d[i] = a[i] & b[i]

        # No null bitmap is created for the result (treat nulls as False)
        out.ptr.null_bitmap = NULL
        return out

    cpdef BoolVector or_vector(self, BoolVector other):
        """Element-wise OR between two BoolVector instances. Returns a new BoolVector."""
        cdef DrakenFixedBuffer* ptr1 = self.ptr
        cdef DrakenFixedBuffer* ptr2 = other.ptr
        cdef Py_ssize_t n = ptr1.length
        if n != ptr2.length:
            raise ValueError("Vectors must have the same length")

        cdef Py_ssize_t nbytes = (n + 7) >> 3
        cdef BoolVector out = BoolVector(<size_t>n)
        cdef uint8_t* a = <uint8_t*> ptr1.data
        cdef uint8_t* b = <uint8_t*> ptr2.data
        cdef uint8_t* d = <uint8_t*> out.ptr.data
        cdef Py_ssize_t i
        for i in range(nbytes):
            d[i] = a[i] | b[i]

        out.ptr.null_bitmap = NULL
        return out

    cpdef BoolVector xor_vector(self, BoolVector other):
        """Element-wise XOR between two BoolVector instances. Returns a new BoolVector."""
        cdef DrakenFixedBuffer* ptr1 = self.ptr
        cdef DrakenFixedBuffer* ptr2 = other.ptr
        cdef Py_ssize_t n = ptr1.length
        if n != ptr2.length:
            raise ValueError("Vectors must have the same length")

        cdef Py_ssize_t nbytes = (n + 7) >> 3
        cdef BoolVector out = BoolVector(<size_t>n)
        cdef uint8_t* a = <uint8_t*> ptr1.data
        cdef uint8_t* b = <uint8_t*> ptr2.data
        cdef uint8_t* d = <uint8_t*> out.ptr.data
        cdef Py_ssize_t i
        for i in range(nbytes):
            d[i] = a[i] ^ b[i]

        out.ptr.null_bitmap = NULL
        return out

    # -------- Ops --------
    cpdef BoolVector take(self, int32_t[::1] indices):
        cdef Py_ssize_t i, n = indices.shape[0]
        cdef BoolVector out = BoolVector(<size_t>n)
        cdef uint8_t* src = <uint8_t*> self.ptr.data
        cdef uint8_t* dst = <uint8_t*> out.ptr.data
        # zero init
        cdef Py_ssize_t out_nbytes = (n + 7) >> 3
        for i in range(out_nbytes):
            dst[i] = 0
        for i in range(n):
            idx = indices[i]
            if ((src[idx >> 3] >> (idx & 7)) & 1) != 0:
                dst[i >> 3] |= (1 << (i & 7))
        return out

    cpdef int8_t[::1] equals(self, bint value):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef Py_ssize_t i, n = ptr.length
        cdef int8_t* buf = <int8_t*> PyMem_Malloc(n)
        if buf == NULL:
            raise MemoryError()
        cdef int target = 1 if value else 0
        for i in range(n):
            val = ((<uint8_t*>ptr.data)[i >> 3] >> (i & 7)) & 1
            buf[i] = 1 if val == target else 0
        return <int8_t[:n]> buf

    cpdef int8_t[::1] not_equals(self, bint value):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef Py_ssize_t i, n = ptr.length
        cdef int8_t* buf = <int8_t*> PyMem_Malloc(n)
        if buf == NULL:
            raise MemoryError()
        cdef int target = 1 if value else 0
        for i in range(n):
            val = ((<uint8_t*>ptr.data)[i >> 3] >> (i & 7)) & 1
            buf[i] = 1 if val != target else 0
        return <int8_t[:n]> buf

    cpdef int8_t any(self):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef Py_ssize_t nbytes = (ptr.length + 7) >> 3
        cdef Py_ssize_t i
        for i in range(nbytes):
            if (<uint8_t*>ptr.data)[i] != 0:
                return 1
        return 0

    cpdef int8_t all(self):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef Py_ssize_t n = ptr.length
        cdef Py_ssize_t i
        for i in range(n):
            if (((<uint8_t*>ptr.data)[i >> 3] >> (i & 7)) & 1) == 0:
                return 0
        return 1

    cpdef int8_t[::1] is_null(self):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef Py_ssize_t i, n = ptr.length
        cdef int8_t* buf = <int8_t*> PyMem_Malloc(n)
        cdef uint8_t byte, bit
        if buf == NULL:
            raise MemoryError()
        if ptr.null_bitmap == NULL:
            for i in range(n):
                buf[i] = 0
        else:
            for i in range(n):
                byte = ptr.null_bitmap[i >> 3]
                bit = (byte >> (i & 7)) & 1
                buf[i] = 0 if bit else 1
        return <int8_t[:n]> buf

    @property
    def null_count(self):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef Py_ssize_t i, n = ptr.length
        cdef Py_ssize_t count = 0
        cdef uint8_t byte, bit
        if ptr.null_bitmap == NULL:
            return 0
        for i in range(n):
            byte = ptr.null_bitmap[i >> 3]
            bit = (byte >> (i & 7)) & 1
            if not bit:
                count += 1
        return count

    cpdef list to_pylist(self):
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef Py_ssize_t i, n = ptr.length
        cdef list out = []
        cdef uint8_t byte, bit
        for i in range(n):
            if ptr.null_bitmap != NULL:
                byte = ptr.null_bitmap[i >> 3]
                bit = (byte >> (i & 7)) & 1
                if not bit:
                    out.append(None)
                    continue
            val = ((<uint8_t*>ptr.data)[i >> 3] >> (i & 7)) & 1
            out.append(bool(val))
        return out

    cdef void hash_into(
        self,
        uint64_t[::1] out_buf,
        Py_ssize_t offset=0
    ) except *:
        cdef DrakenFixedBuffer* ptr = self.ptr
        cdef Py_ssize_t n = ptr.length
        if n == 0:
            return

        if offset < 0 or offset + n > out_buf.shape[0]:
            raise ValueError("BoolVector.hash_into: output buffer too small")

        cdef Py_ssize_t i
        cdef Py_ssize_t block = 0
        cdef Py_ssize_t j = 0
        cdef Py_ssize_t idx = 0
        cdef uint8_t byte, bit
        cdef uint64_t value
        cdef uint64_t* dst = &out_buf[offset]
        cdef uint8_t* values = <uint8_t*> ptr.data
        cdef bint has_nulls = ptr.null_bitmap != NULL
        cdef uint64_t[BOOL_HASH_CHUNK] scratch
        cdef uint64_t* scratch_ptr = <uint64_t*> scratch

        if not has_nulls:
            i = 0
            while i < n:
                block = n - i
                if block > BOOL_HASH_CHUNK:
                    block = BOOL_HASH_CHUNK
                for j in range(block):
                    idx = i + j
                    if (values[idx >> 3] >> (idx & 7)) & 1:
                        scratch[j] = TRUE_HASH
                    else:
                        scratch[j] = FALSE_HASH
                simd_mix_hash(dst + i, scratch_ptr, <size_t> block)
                i += block
            return

        for i in range(n):
            byte = ptr.null_bitmap[i >> 3]
            bit = (byte >> (i & 7)) & 1
            if not bit:
                value = NULL_HASH
            else:
                if ((values)[i >> 3] >> (i & 7)) & 1:
                    value = TRUE_HASH
                else:
                    value = FALSE_HASH

            dst[i] = mix_hash(dst[i], value)

    def __str__(self):
        cdef list vals = []
        cdef Py_ssize_t i, k = min(<Py_ssize_t>buf_length(self.ptr), 10)
        for i in range(k):
            vals.append(bool(((<uint8_t*>self.ptr.data)[i >> 3] >> (i & 7)) & 1))
        return f"<BoolVector len={buf_length(self.ptr)} values={vals}>"


cdef BoolVector from_arrow(object array):
    cdef BoolVector vec = BoolVector(0, True)
    vec.ptr = <DrakenFixedBuffer*> malloc(sizeof(DrakenFixedBuffer))
    if vec.ptr == NULL:
        raise MemoryError()
    vec.owns_data = False

    cdef object bufs = array.buffers()
    # Keep references to prevent GC
    vec._arrow_null_buf = bufs[0]
    vec._arrow_data_buf = bufs[1]

    cdef intptr_t base_ptr = <intptr_t> bufs[1].address
    cdef Py_ssize_t offset = array.offset
    cdef intptr_t nb_addr

    vec.ptr.type = DRAKEN_BOOL
    vec.ptr.itemsize = 1
    vec.ptr.length = <size_t> len(array)

    # Data buffer handling
    cdef Py_ssize_t nbytes
    cdef uint8_t* src_data
    cdef uint8_t* dst_data
    cdef object new_data_bytes
    cdef Py_ssize_t i

    if offset % 8 == 0:
        # Aligned offset: zero-copy
        vec.ptr.data = <void*> (base_ptr + (offset >> 3))
    else:
        # Unaligned offset: must copy and shift data
        nbytes = (len(array) + 7) // 8
        new_data_bytes = PyBytes_FromStringAndSize(NULL, nbytes)
        dst_data = <uint8_t*> PyBytes_AS_STRING(new_data_bytes)
        memset(dst_data, 0, nbytes)
        
        src_data = <uint8_t*> base_ptr
        
        # Copy bits shifting them
        for i in range(len(array)):
            if (src_data[(offset + i) >> 3] >> ((offset + i) & 7)) & 1:
                dst_data[i >> 3] |= (1 << (i & 7))
        
        vec.ptr.data = <void*> dst_data
        vec._arrow_data_buf = new_data_bytes

    # Null bitmap handling
    cdef uint8_t* src_bitmap
    cdef uint8_t* dst_bitmap
    cdef object new_bitmap_bytes

    if bufs[0] is not None:
        nb_addr = bufs[0].address
        
        if offset % 8 == 0:
            vec.ptr.null_bitmap = (<uint8_t*> nb_addr) + (offset >> 3)
        else:
            # Unaligned offset: must copy and shift nulls
            nbytes = (len(array) + 7) // 8
            new_bitmap_bytes = PyBytes_FromStringAndSize(NULL, nbytes)
            dst_bitmap = <uint8_t*> PyBytes_AS_STRING(new_bitmap_bytes)
            memset(dst_bitmap, 0, nbytes)
            
            src_bitmap = <uint8_t*> nb_addr
            
            for i in range(len(array)):
                if (src_bitmap[(offset + i) >> 3] >> ((offset + i) & 7)) & 1:
                    dst_bitmap[i >> 3] |= (1 << (i & 7))
            
            vec.ptr.null_bitmap = dst_bitmap
            vec._arrow_null_buf = new_bitmap_bytes
    else:
        vec.ptr.null_bitmap = NULL

    return vec


cdef BoolVector from_sequence(uint8_t[::1] data):
    """
    Create BoolVector from a typed uint8 memoryview (zero-copy, bit-packed).
    
    Args:
        data: uint8_t[::1] memoryview (C-contiguous, bit-packed: 8 bools per byte)
    
    Returns:
        BoolVector wrapping the memoryview data
    
    Note:
        Input data should be bit-packed (8 boolean values per byte).
        The length will be inferred as data.shape[0] * 8.
    """
    cdef BoolVector vec = BoolVector(0, True)
    vec.ptr = <DrakenFixedBuffer*> malloc(sizeof(DrakenFixedBuffer))
    if vec.ptr == NULL:
        raise MemoryError()
    vec.owns_data = False

    # Keep reference to prevent GC
    vec._arrow_data_buf = data.base if data.base is not None else data
    vec._arrow_null_buf = None

    vec.ptr.type = DRAKEN_BOOL
    vec.ptr.itemsize = 1
    # Bit-packed: 8 booleans per byte
    vec.ptr.length = <size_t> (data.shape[0] * 8)
    vec.ptr.data = <void*> &data[0]
    vec.ptr.null_bitmap = NULL

    return vec



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.


from collections import defaultdict


class _QueryStatistics:
    def __init__(self):
        # predefine "messages" and "executed_plan" so all new statistics default to 0
        self._stats: dict = defaultdict(int)
        self._stats["messages"] = []
        self._stats["executed_plan"] = None

    def _ns_to_s(self, nano_seconds: int) -> float:
        """convert elapsed ns to s"""
        if nano_seconds == 0:
            return 0
        return nano_seconds / 1e9

    def __getattr__(self, attr):
        """allow access using stats.statistic_name"""
        return self._stats[attr]

    def __setattr__(self, attr, value):
        """allow access using stats.statistic_name"""
        if attr == "_stats":
            super().__setattr__(attr, value)
        else:
            self._stats[attr] = value

    def increase(self, attr: str, amount: float = 1.0):
        self._stats[attr] += amount

    def add_message(self, message: str):
        """collect warnings"""
        self._stats["messages"].append(message)

    def as_dict(self):
        """
        Return statistics as a dictionary
        """
        import opteryx

        stats_dict = dict(self._stats)
        for k, v in stats_dict.items():
            # times are recorded in ns but reported in seconds
            if k.startswith("time_"):
                stats_dict[k] = self._ns_to_s(v)
        stats_dict["time_total"] = self._ns_to_s(
            stats_dict.pop("end_time", 0) - stats_dict.pop("start_time", 0)
        )
        # sort the keys in the dictionary
        stats_dict = {key: stats_dict[key] for key in sorted(stats_dict)}
        # put messages and executed_plan at the end
        stats_dict["version"] = opteryx.__version__
        stats_dict["messages"] = stats_dict.pop("messages", [])
        stats_dict["executed_plan"] = stats_dict.pop("executed_plan", None)
        return stats_dict


class QueryStatistics(_QueryStatistics):
    slots = "_instances"

    _instances: dict[str, _QueryStatistics] = {}

    def __new__(cls, qid=""):
        if cls._instances.get(qid) is None:
            cls._instances[qid] = _QueryStatistics()
            if len(cls._instances.keys()) > 16:
                # find the first key that is not "system"
                key_to_remove = next((key for key in cls._instances if key != "system"), None)
                if key_to_remove:
                    cls._instances.pop(key_to_remove)
        return cls._instances[qid]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

from typing import Any
from typing import Dict
from typing import Optional

import numpy
from orso.schema import RelationSchema
from orso.types import OrsoTypes

from opteryx.compiled.structures.relation_statistics import to_int
from opteryx.managers.expression import NodeType
from opteryx.models import RelationStatistics
from opteryx.shared.stats_cache import StatsCache
from opteryx.third_party.cyan4973.xxhash import hash_bytes

handlers = {
    "Eq": lambda v, min_, max_: v < min_ or v > max_,
    "NotEq": lambda v, min_, max_: min_ == max_ == v,
    "Gt": lambda v, min_, max_: max_ < v,
    "GtEq": lambda v, min_, max_: max_ <= v,
    "Lt": lambda v, min_, max_: min_ > v,
    "LtEq": lambda v, min_, max_: min_ >= v,
}


class Statistics:
    def __init__(self, statistics: dict, **kwargs):
        self.stats_cache = StatsCache()
        self.relation_statistics = RelationStatistics()

    def read_blob_statistics(
        self, blob_name: str, blob_bytes: bytes = None, decoder=None
    ) -> Optional[Dict[str, Any]]:
        key = hex(hash_bytes(blob_name.encode())).encode()
        cached_stats = self.stats_cache.get(key)
        if cached_stats is not None:
            # If statistics are cached, return them
            return cached_stats

        cached_stats = decoder(blob_bytes, just_statistics=True)
        if cached_stats is not None:
            self.stats_cache.set(key, cached_stats)
        return cached_stats

    def prune_blobs(self, blob_names: list[str], query_statistics, selection) -> list[str]:
        new_blob_names = []

        for blob_name in blob_names:
            key = hex(hash_bytes(blob_name.encode())).encode()
            cached_stats = self.stats_cache.get(key)

            skip_blob = False

            # if we have no stats we can't make a decision
            if cached_stats is not None:
                valid_conditions = [
                    cond
                    for cond in selection
                    if cond.value in handlers
                    and cond.left.node_type == NodeType.IDENTIFIER
                    and cond.right.node_type == NodeType.LITERAL
                    and cond.left.schema_column.type
                    not in (OrsoTypes.DATE, OrsoTypes.TIME, OrsoTypes.TIMESTAMP)
                    and cond.right.schema_column.type
                    not in (OrsoTypes.DATE, OrsoTypes.TIME, OrsoTypes.TIMESTAMP)
                ]
                # Handle AnyOp* comparisons where left is a literal and right is an identifier
                valid_conditions_any = [
                    cond
                    for cond in selection
                    if cond.value == "AnyOpEq"
                    and cond.left.node_type == NodeType.LITERAL
                    and cond.right.node_type == NodeType.IDENTIFIER
                    and cond.right.schema_column.type
                    not in (OrsoTypes.DATE, OrsoTypes.TIME, OrsoTypes.TIMESTAMP)
                    and cond.left.schema_column.type
                    not in (OrsoTypes.DATE, OrsoTypes.TIME, OrsoTypes.TIMESTAMP)
                ]

                for condition in valid_conditions:
                    column_name = condition.left.source_column.encode()
                    literal_value = condition.right.value
                    if type(literal_value) is numpy.datetime64:
                        literal_value = str(literal_value.astype("M8[ms]"))
                    if hasattr(literal_value, "item"):
                        literal_value = literal_value.item()
                    literal_value = to_int(literal_value)
                    max_value = cached_stats.upper_bounds.get(column_name)
                    min_value = cached_stats.lower_bounds.get(column_name)

                    if max_value is not None and min_value is not None:
                        prune = handlers.get(condition.value)
                        if prune and prune(literal_value, min_value, max_value):
                            query_statistics.blobs_pruned += 1
                            skip_blob = True
                            break

                # Evaluate AnyOp* conditions (literal = ANY(column)) safely using element min/max
                for condition in valid_conditions_any:
                    column_name = condition.right.source_column.encode()
                    literal_value = condition.left.value
                    # Skip NULL literals — unsafe to prune
                    if literal_value is None:
                        continue
                    if type(literal_value) is numpy.datetime64:
                        # convert to python datetime for consistent to_int conversion
                        literal_value = literal_value.astype("M8[ms]").astype("O")
                    if hasattr(literal_value, "item"):
                        literal_value = literal_value.item()
                    literal_value = to_int(literal_value)
                    # Skip NULL_FLAG values (NaN/unconvertible) that appear as the NULL sentinel
                    NULL_FLAG = -(1 << 63)
                    if literal_value == NULL_FLAG:
                        continue
                    max_value = cached_stats.upper_bounds.get(column_name)
                    min_value = cached_stats.lower_bounds.get(column_name)

                    if max_value is not None and min_value is not None:
                        # convert AnyOpEq -> Eq, AnyOpGt -> Gt, etc.
                        op_name = condition.value[5:]
                        prune = handlers.get(op_name)
                        if prune and prune(literal_value, min_value, max_value):
                            query_statistics.blobs_pruned += 1
                            skip_blob = True
                            break

            if not skip_blob:
                new_blob_names.append(blob_name)

        return new_blob_names

    def map_statistics(
        self, statistics: Optional[RelationStatistics], schema: RelationSchema
    ) -> RelationSchema:
        if statistics is None:
            return schema

        schema.row_count_metric = statistics.record_count

        for column in schema.columns:
            # Statistics dictionaries use bytes keys, so we need to encode the column name
            column_key = column.name.encode() if isinstance(column.name, str) else column.name
            column.highest_value = statistics.upper_bounds.get(column_key, None)
            column.lowest_value = statistics.lower_bounds.get(column_key, None)
            if statistics.null_count:
                column.null_count = statistics.null_count.get(column_key, None)

        return schema



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Set Variables Node

This is a SQL Query Execution Plan Node.
"""

from opteryx.constants import QueryStatus
from opteryx.models import NonTabularResult
from opteryx.models import QueryProperties

from . import BasePlanNode


class SetVariableNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **parameters):
        BasePlanNode.__init__(self, properties=properties, **parameters)

        self.variable = parameters.get("variable")
        self.value = parameters.get("value")
        self.variables = parameters.get("variables")

    @property
    def name(self):  # pragma: no cover
        return "Set Variables"

    @property
    def config(self):  # pragma: no cover
        return f"{self.variable} TO {self.value}"

    def __call__(self, morsel, **kwargs) -> NonTabularResult:
        self.variables[self.variable] = self.value
        return NonTabularResult(record_count=1, status=QueryStatus.SQL_SUCCESS)  # type: ignore



# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

from libc.stdint cimport uint64_t, int64_t
from libc.stdlib cimport malloc, free

from opteryx.compiled.structures.hash_table cimport HashTable
from opteryx.compiled.table_ops.hash_ops cimport compute_row_hashes
from opteryx.compiled.table_ops.null_avoidant_ops cimport non_null_row_indices
from opteryx.utils.arrow import align_tables

import numpy
cimport numpy
numpy.import_array()

cpdef HashTable probe_side_hash_map(object relation, list join_columns):
    """
    Build a hash table for the join operations (probe-side) using buffer-level hashing.
    """
    cdef HashTable ht = HashTable()
    cdef int64_t num_rows = relation.num_rows
    cdef int64_t[::1] non_null_indices
    cdef uint64_t[::1] row_hashes = numpy.empty(num_rows, dtype=numpy.uint64)
    cdef Py_ssize_t i

    non_null_indices = non_null_row_indices(relation, join_columns)

    # Compute hash of each row on the buffer level
    compute_row_hashes(relation, join_columns, row_hashes)

    # Insert into HashTable using row index + buffer-computed hash
    for i in range(non_null_indices.shape[0]):
        ht.insert(row_hashes[non_null_indices[i]], non_null_indices[i])

    return ht


def right_join(
    left_relation,
    right_relation,
    left_columns: List[str],
    right_columns: List[str],
    left_hash,
    filter_index,
    columns=None
):
    """
    Perform a RIGHT JOIN.

    This implementation ensures that all rows from the right table are included in the result set,
    with rows from the left table matched where possible, and columns from the left table
    filled with NULLs where no match is found.

    Parameters:
        left_relation (pyarrow.Table): The left pyarrow.Table to join.
        right_relation (pyarrow.Table): The right pyarrow.Table to join.
        left_columns (list of str): Column names from the left table to join on.
        right_columns (list of str): Column names from the right table to join on.
        columns (list of str, optional): Columns to include in the result. If None, all columns are included.

    Yields:
        pyarrow.Table: A chunk of the result of the RIGHT JOIN operation.
    """
    # Build hash table of left side
    cdef HashTable left_hash_table = HashTable()
    cdef Py_ssize_t num_left_rows = left_relation.num_rows
    cdef Py_ssize_t i

    cdef uint64_t* raw_hashes = <uint64_t*> malloc(num_left_rows * sizeof(uint64_t))
    if raw_hashes == NULL:
        raise MemoryError("Failed to allocate memory for hash table")
    cdef uint64_t[::1] left_hashes = <uint64_t[:num_left_rows]> raw_hashes

    compute_row_hashes(left_relation, left_columns, left_hashes)

    for i in range(num_left_rows):
        left_hash_table.insert(left_hashes[i], i)

    free(raw_hashes)

    cdef uint64_t* chunk_hashes
    cdef uint64_t[::1] right_hashes

    # Iterate over the right_relation in chunks
    for right_chunk in right_relation.to_batches(50_000):
        chunk_size = right_chunk.num_rows

        # Compute hashes for this right chunk
        chunk_hashes = <uint64_t*> malloc(chunk_size * sizeof(uint64_t))
        if chunk_hashes == NULL:
            raise MemoryError("Failed to allocate memory for chunk hashes")
        right_hashes = <uint64_t[:chunk_size]> chunk_hashes

        compute_row_hashes(right_chunk, right_columns, right_hashes)

        # Collect matches
        left_indexes = []
        right_indexes = []

        for i in range(chunk_size):
            left_matches = left_hash_table.get(right_hashes[i])
            if left_matches.size() > 0:
                left_indexes.extend(left_matches)
                right_indexes.extend([i] * len(left_matches))
            else:
                left_indexes.append(None)
                right_indexes.append(i)

        free(chunk_hashes)

        yield align_tables(left_relation, right_chunk, left_indexes, right_indexes)
